import datetime
from datetime import timedelta, timezone
from numbers import Number

import numpy as np
from scipy.signal import argrelextrema


def _parse_args(args, kwargs, names, *defaults):
    assert len(defaults) == len(names)
    result = list(args) + [None] * (len(names) - len(args))
    for i, (name, res) in enumerate(zip(names, result)):
        result[i] = kwargs.get(name, res if res is not None else defaults[i])
    return result


def r2(true, pred):
    """Coefficient of determination (R²)"""
    nobs = len(pred)
    true = true[-nobs:]
    return 1 - np.sum((true - pred)**2) / np.sum((true - np.mean(true))**2)


def rmse(true, pred):
    """Root mean squared error"""
    nobs = len(pred)
    return np.sqrt(np.sum((true[-nobs:] - pred) ** 2) / nobs)


def mape(true, pred):
    """Mean absolute percentage error"""
    nobs = len(pred)
    return np.mean(np.abs(true[-nobs:] - pred)) / np.abs(true).mean()


def mae(true, pred):
    """Median absolute error"""
    nobs = len(pred)
    return np.median(np.abs(true[-nobs:] - pred))


def pocid(true, pred):
    """Prediction on change of direction"""
    nobs = len(pred)
    return 100 * np.mean((np.diff(true[-nobs:]) * np.diff(pred)) > 0)


def _detrend(x, type):
    if type == 'diff':
        x = np.diff(x)
    elif isinstance(type, str):
        type = dict(constant=0, linear=1, quadratic=2, cubic=3)[type]
    if isinstance(type, Number):
        import statsmodels.api as sm
        x = sm.tsa.detrend(x, type)
    return x


def _significant_periods(periods, pgram):
    if pgram.size == 0:
        return periods, pgram

    # Order ascending
    periods = periods[::-1]
    pgram = pgram[::-1]
    # Scale and extract significant
    pgram = (pgram - pgram.min()) / pgram.ptp()
    significant = argrelextrema(pgram, np.greater, order=5)
    return periods[significant], pgram[significant]


def periodogram(x, *args, detrend='diff', **kwargs):
    """
    Return periodogram of signal `x`.

    Parameters
    ----------
    x: array_like
        A 1D signal.
    detrend: 'diff' or False or int
        Remove trend from x. If int, fit and subtract a polynomial of this
        order. See also: `statsmodels.tsa.detrend`.
    args, kwargs:
        As accepted by `scipy.signal.periodogram`.

    Returns
    -------
    periods: array_like
        The periods at which the spectral density is calculated.
    pgram: array_like
        Power spectral density of x.
    """
    from scipy.signal import periodogram
    x = _detrend(x, detrend)
    freqs, pgram = periodogram(x, *args, detrend=False, **kwargs)

    SKIP = len(x) // 1000  # HACK: For long series, the first few frequency/period values are "unstable".
    freqs, pgram = freqs[SKIP:], pgram[SKIP:]

    periods = 1 / freqs
    periods, pgram = _significant_periods(periods, pgram)
    return periods, pgram


def periodogram_nonequispaced(times, x, *, freqs=None,
                              period_low=None, period_high=None,
                              n_periods=1000, detrend='linear'):
    """
    Compute the Lomb-Scargle periodogram for non-equispaced timeseries.

    Parameters
    ----------
    times: array_like
        Sample times.
    x: array_like
        A 1D signal.
    freqs: array_like, optional
        **Angular** frequencies for output periodogram.
    period_low: float
        If `freqs` not provided, the lowest period for which to look for
        periodicity. Defaults to 5th percentile of time difference between
        observations.
    period_high: float
        If `freqs` not provided, the highest period for which to look for
        periodicity. Defaults to 80th percentile of time difference of
        observations, or 200*period_low, whichever is larger.
    n_periods: int
        Number of periods between period_low and period_high to try.
    detrend: 'diff' or False or int
        Remove trend from x. If int, fit and subtract a polynomial of this
        order. See also: `statsmodels.tsa.detrend`.

    Returns
    -------
    periods: array_like
        The periods at which the spectral density is calculated.
    pgram: array_like
        Lomb-Scargle periodogram.

    Notes
    -----
    Read also:
    https://jakevdp.github.io/blog/2015/06/13/lomb-scargle-in-python/#lomb-scargle-algorithms-in-python
    """
    from scipy.signal import lombscargle
    x = _detrend(x, detrend)
    if detrend == 'diff':
        times = times[1:]

    if freqs is None:
        percentile = np.percentile(np.diff(times), [5, 80])
        if period_low is None:
            period_low = percentile[0]
        if period_high is None:
            period_high = max(200 * period_low, percentile[1])
        # Periods *from high to low* because they are reversed later!
        periods = np.linspace(period_high, period_low, n_periods)
        freqs = 2 * np.pi / periods
    else:
        periods = 2 * np.pi / freqs

    pgram = lombscargle(times, x, freqs)
    # Normalize -- I have no idea what I am doing; took this from
    # https://jakevdp.github.io/blog/2015/06/13/lomb-scargle-in-python/#lomb-scargle-algorithms-in-python
    pgram *= 2 / (len(x) * x.std()**2)

    periods, pgram = _significant_periods(periods, pgram)
    return periods, pgram


def _significant_acf(corr, has_confint):
    if has_confint:
        corr, confint = corr

    periods = argrelextrema(np.abs(corr), np.greater, order=3)[0]
    corr = corr[periods]
    if has_confint:
        confint = confint[periods]

    result = np.column_stack((periods, corr))
    if has_confint:
        result = (result, np.column_stack((periods, confint)))
    return result


def autocorrelation(x, *args, unbiased=True, nlags=None, fft=True, **kwargs):
    """
    Return autocorrelation function of signal `x`.

    Parameters
    ----------
    x: array_like
        A 1D signal.
    nlags: int
        The number of lags to calculate the correlation for (default .9*len(x))
    fft:  bool
        Compute the ACF via FFT.
    args, kwargs
        As accepted by `statsmodels.tsa.stattools.acf`.

    Returns
    -------
    acf: array
        Autocorrelation function.
    confint: array, optional
        Confidence intervals if alpha kwarg provided.
    """
    from statsmodels.tsa.stattools import acf
    if nlags is None:
        nlags = int(.9 * len(x))
    corr = acf(x, *args, unbiased=unbiased, nlags=nlags, fft=fft, **kwargs)
    return _significant_acf(corr, kwargs.get('alpha'))


def partial_autocorrelation(x, *args, nlags=None, method='ldb', **kwargs):
    """
    Return partial autocorrelation function (PACF) of signal `x`.

    Parameters
    ----------
    x: array_like
        A 1D signal.
    nlags: int
        The number of lags to calculate the correlation for
        (default: min(len(x)//2 - 1, len(x) - 1))
    args, kwargs
        As accepted by `statsmodels.tsa.stattools.pacf`.

    Returns
    -------
    acf: array
        Partioal autocorrelation function.
    confint : optional
        As returned by `statsmodels.tsa.stattools.pacf`.
    """
    from statsmodels.tsa.stattools import pacf
    if nlags is None:
        nlags = min(len(x)//2 - 1, len(x) - 1)
    corr = pacf(x, *args, nlags=nlags, method=method, **kwargs)
    return _significant_acf(corr, kwargs.get('alpha'))


def interpolate_timeseries(data, method='linear', multivariate=False):
    """Return a new Timeseries (Table) with nan values interpolated.

    Parameters
    ----------
    data : Orange.data.Table
        A table to interpolate.
    method : str {'linear', 'cubic', 'nearest', 'mean'}
        The interpolation method to use.
    multivariate : bool
        Whether to perform multivariate (2d) interpolation first.
        Univariate interpolation of same method is always performed as a
        final step.

    Returns
    -------
    series : Timeseries
        A table with nans in original replaced with interpolated values.
    """
    from scipy.interpolate import griddata, interp1d
    from Orange.data import Domain
    from orangecontrib.timeseries import Timeseries

    attrs = data.domain.attributes
    cvars = data.domain.class_vars
    metas = data.domain.metas
    X = data.X.copy()
    Y = np.column_stack((data.Y,)).copy()  # make 2d
    M = data.metas.copy()

    # Interpolate discrete columns to mode/nearest value
    _x = Timeseries.from_data_table(data).time_values.astype(float)
    for A, vars in ((X, attrs),
                    (Y, cvars)):
        for i, var in enumerate(vars):
            if not var.is_discrete:
                continue
            vals = A[:, i]
            isnan = np.isnan(vals)
            if not isnan.any():
                continue
            if method == 'nearest':
                nonnan = ~isnan
                x, vals = _x[nonnan], vals[nonnan]
                f = interp1d(x, vals, kind='nearest', copy=False, assume_sorted=True)
                A[isnan, i] = f(_x)[isnan]
                continue
            A[isnan, i] = np.argmax(np.bincount(vals[~isnan].astype(int)))

    # Interpolate data
    if multivariate and method != 'mean':
        for A, vars in ((X, attrs),
                        (Y, cvars)):
            is_continuous = [var.is_continuous for var in vars]
            if sum(is_continuous) < 3 or A.shape[0] < 3:
                # griddata() doesn't work with 1d data
                continue

            # Only multivariate continuous features
            Acont = A[:, is_continuous]
            isnan = np.isnan(Acont)
            if not isnan.any():
                continue
            nonnan = ~isnan
            vals = griddata(nonnan.nonzero(), Acont[nonnan], isnan.nonzero(),
                            method=method)
            Acont[isnan] = vals
            A[:, is_continuous] = Acont

    # Do the 1d interpolation anyway in case 2d left some nans
    for A in (X, Y):
        for i, col in enumerate(A.T):
            isnan = np.isnan(col)
            # there is no need to interpolate if there are no nans
            # there needs to be at least two numbers
            if not isnan.any() or sum(~isnan) < 2:
                continue

            # Mean interpolation
            if method == 'mean':
                A[isnan, i] = np.nanmean(col)
                continue

            nonnan = ~isnan
            f = interp1d(_x[nonnan], col[nonnan], kind=method,
                         copy=False, assume_sorted=True, bounds_error=False)
            A[isnan, i] = f(_x[isnan])

            # nearest-interpolate any nans at vals start and end
            # TODO: replace nearest with linear/OLS?
            valid = (~np.isnan(col)).nonzero()[0]
            first, last = valid[0], valid[-1]
            col[:first] = col[first]
            col[last:] = col[last]

    ts = Timeseries.from_numpy(Domain(attrs, cvars, metas), X, Y, M)
    return ts


def seasonal_decompose(data, model='multiplicative', period=12, *, callback=None):
    """
    Return table of decomposition components of original features and
    original features seasonally adjusted.

    Parameters
    ----------
    data : Timeseries
        A table of featres to decompose/adjust.
    model : str {'additive', 'multiplicative'}
        A decompostition model. See:
        https://en.wikipedia.org/wiki/Decomposition_of_time_series
    period : int
        The period length of season.
    callback : callable
        Optional callback to call (with no parameters) after each iteration.

    Returns
    -------
    table : Timeseries
        Table with columns: original series seasonally adjusted, original
        series' seasonal components, trend components, and residual components.
    """
    from operator import sub, truediv
    from Orange.data import Domain, ContinuousVariable
    from orangecontrib.timeseries import Timeseries
    from orangecontrib.timeseries.widgets.utils import available_name
    import statsmodels.api as sm

    def _interp_trend(trend):
        first = next(i for i, val in enumerate(trend) if val == val)
        last = trend.size - 1 - next(
            i for i, val in enumerate(trend[::-1]) if val == val)
        d = 3
        first_last = min(first + d, last)
        last_first = max(first, last - d)

        k, n = np.linalg.lstsq(
            np.column_stack((np.arange(first, first_last), np.ones(first_last - first))),
            trend[first:first_last])[0]
        trend[:first] = np.arange(0, first) * k + n

        k, n = np.linalg.lstsq(
            np.column_stack((np.arange(last_first, last), np.ones(last - last_first))),
            trend[last_first:last])[0]
        trend[last + 1:] = np.arange(last + 1, trend.size) * k + n
        return trend

    attrs = []
    X = []
    recomposition = sub if model == 'additive' else truediv
    interp_data = data.interp()
    for var in data.domain.variables:
        decomposed = sm.tsa.seasonal_decompose(np.ravel(interp_data[:, var]),
                                               model=model,
                                               freq=period)
        adjusted = recomposition(decomposed.observed,
                                 decomposed.seasonal)

        season = decomposed.seasonal
        trend = _interp_trend(decomposed.trend)
        resid = recomposition(adjusted, trend)

        # Re-apply nans
        isnan = np.isnan(data[:, var]).ravel()
        adjusted[isnan] = np.nan
        trend[isnan] = np.nan
        resid[isnan] = np.nan

        attrs.extend(
            ContinuousVariable(
                available_name(data.domain,
                               var.name + ' ({})'.format(transform)))
            for transform in
            ('season. adj.', 'seasonal', 'trend', 'residual')
        )
        X.extend((adjusted, season, trend, resid))

        if callback:
            callback()

    ts = Timeseries.from_numpy(Domain(attrs), np.column_stack(X))
    return ts


def granger_causality(data, max_lag=10, alpha=.05, *, callback=None):
    """
    Return results of Granger-causality tests.

    Parameters
    ----------
    data : Timeseries
        A table of features to compute Granger causality between.
    max_lag : int
        The maximum lag to compute Granger-causality for.
    alpha : float in (0, 1)
        Confidence of test is 1 - alpha.
    callback : callable
        A callback to call in each iteration with ratio of completion.

    Returns
    -------
    res : list of lists
        Each internal list is [lag, p-value, antecedent, consequent] where
        lag is the minimum lag at which antecedent feature in data is
        Granger-causal for the consequent feature in data.
    """
    from statsmodels.tsa.stattools import grangercausalitytests
    from Orange.data import Table, Domain
    # TODO: use VAR Granger causality
    # TODO: consider CCM in stead/addition of GC: https://en.wikipedia.org/wiki/Convergent_cross_mapping
    # http://statsmodels.sourceforge.net/devel/generated/statsmodels.tsa.vector_ar.var_model.VARResults.test_causality.html
    # http://statsmodels.sourceforge.net/devel/vector_ar.html#granger-causality

    data = data.interp()
    domain = [var for var in data.domain.variables if var.is_continuous]
    res = []

    step = 0
    for row_attr in domain:
        for col_attr in domain:
            if row_attr == col_attr or data.time_variable in (row_attr, col_attr):
                continue
            X = Table(Domain([], [], [col_attr, row_attr], data.domain), data).metas
            try:
                tests = grangercausalitytests(X, max_lag, verbose=False)
                lag, p_val = next(
                    (
                        (lag, tests[lag][0]["ssr_ftest"][1])
                        for lag in range(1, 1 + max_lag)
                        if tests[lag][0]["ssr_ftest"][1] < alpha
                    ),
                    (None, None),
                )
            except ValueError:
                lag = p_val = None
            if lag:
                res.append([lag, p_val, row_attr.name, col_attr.name])

            step += 1
            if callback:
                callback(step / ((len(domain) - 1) ** 2 - len(domain) + 1))
    return res


def moving_transform(data, spec, fixed_wlen=0):
    """
    Return data transformed according to spec.

    Parameters
    ----------
    data : Timeseries
        A table with features to transform.
    spec : list of lists
        A list of lists [feature:Variable, window_length:int, function:callable].
    fixed_wlen : int
        If not 0, then window_length in spec is disregarded and this length
        is used. Also the windows don't shift by one but instead align
        themselves side by side.

    Returns
    -------
    transformed : Timeseries
        A table of original data its transformations.
    """
    from itertools import chain
    from Orange.data import ContinuousVariable, Domain
    from orangecontrib.timeseries import Timeseries
    from orangecontrib.timeseries.widgets.utils import available_name
    from orangecontrib.timeseries.agg_funcs import Cumulative_sum, Cumulative_product

    X = []
    attrs = []

    for var, wlen, func in spec:
        col = np.ravel(data[:, var])

        if fixed_wlen:
            wlen = fixed_wlen

        if func in (Cumulative_sum, Cumulative_product):
            out = list(chain.from_iterable(func(col[i:i + wlen])
                                           for i in range(0, len(col), wlen)))
        else:
            # In reverse cause lazy brain. Also prefer informative ends, not beginnings as much
            col = col[::-1]
            out = [func(col[i:i + wlen])
                   for i in range(0, len(col), wlen if bool(fixed_wlen) else 1)]
            out = out[::-1]

        X.append(out)

        template = '{} ({}; {})'.format(var.name, wlen, func.__name__.lower().replace('_', ' '))
        name = available_name(data.domain, template)
        attrs.append(ContinuousVariable(name))

    dataX, dataY, dataM = data.X, data.Y, data.metas
    if fixed_wlen:
        n = len(X[0])
        dataX = dataX[::-1][::fixed_wlen][:n][::-1]
        dataY = dataY[::-1][::fixed_wlen][:n][::-1]
        dataM = dataM[::-1][::fixed_wlen][:n][::-1]

    ts = Timeseries.from_numpy(Domain(data.domain.attributes + tuple(attrs),
                                      data.domain.class_vars,
                                      data.domain.metas),
                               np.column_stack(
                                   (dataX, np.column_stack(X))) if X else dataX,
                               dataY, dataM)
    ts.time_variable = data.time_variable
    return ts


def model_evaluation(data, models, n_folds, forecast_steps, *, callback=None):
    """
    Evaluate models on data.

    Parameters
    ----------
    data : Timeseries
        The timeseries to model. Must have a class variable that is used
        for prediction and scoring.
    models : list
        List of models (objects with fit() and predict() methods) to try.
    n_folds : int
        Number of iterations.
    forecast_steps : int
        Number of forecast steps at each iteraction.
    callback : callable, optional
        Optional argument-less callback to call after each iteration.

    Returns
    -------
    results : list of lists
        A table with horizontal and vertical headers and results. Print it
        to see it.
    """
    if not data.domain.class_var:
        raise ValueError('Data requires a target variable. Use Select Columns '
                         'widget to set one variable as target.')
    max_lag = max(m.max_order for m in models)
    if n_folds * forecast_steps + max_lag > len(data):
        raise ValueError(
            'Supplied time series is too short for this many folds '
            '/ step size. Retry with fewer iterations.')

    def _score_vector(model, true, pred):
        true = np.asanyarray(true)
        pred = np.asanyarray(pred)
        nonnan = ~np.isnan(true)
        if not nonnan.all():
            pred = pred[nonnan]
            true = true[nonnan]
        row = [str(getattr(model, 'name', model))]
        if pred.size:
            row.extend(score(true, pred) for score in (rmse, mae, mape, pocid, r2))
        else:
            row.extend(['err'] * 5)
        try:
            row.extend([model.results.aic, model.results.bic])
        except Exception:
            row.extend(['err'] * 2)
        return row

    res = [['Model', 'RMSE', 'MAE', 'MAPE', 'POCID', 'R²', 'AIC', 'BIC']]
    interp_data = data.interp()
    true_y = np.ravel(data[:, data.domain.class_var])

    for model in models:
        full_true = []
        full_pred = []
        for fold in range(1, n_folds + 1):
            train_end = -fold * forecast_steps
            try:
                model.fit(interp_data[:train_end])
                pred, _, _ = model.predict(forecast_steps)
            except Exception:
                continue
            finally:
                if callback:
                    callback()

            full_true.extend(true_y[train_end:][:forecast_steps])  # Sliced twice because it doesn't work at the end, e.g. [-3:0] == [] :(
            full_pred.extend(np.c_[pred][:, 0])  # Only interested in the class var
            assert len(full_true) == len(full_pred)

        res.append(_score_vector(model, full_true, full_pred))

        # Score in-sample fittedvalues
        try:
            model.fit(interp_data)
            fittedvalues = model.fittedvalues()
            if fittedvalues.ndim > 1:
                fittedvalues = fittedvalues[..., 0]
        except Exception:
            row = ['err'] * 7
        else:
            row = _score_vector(model, true_y, fittedvalues)
        row[0] = row[0] + ' (in-sample)'
        res.append(row)
    return res


SECONDS = 126230400
DAYS = 1461


def timestamp(dt):
    try:
        ts = dt.timestamp()
    except OverflowError:
        if not dt.tzinfo:
            # treat datetime as in local timezone
            dt = dt.astimezone()
        # compute timestamp manually
        ts = (dt - datetime.datetime(1970, 1, 1, tzinfo=timezone.utc)) /\
             timedelta(seconds=1)
    return ts


def fromtimestamp(ts, tz=None):
    try:
        dt = datetime.datetime.fromtimestamp(ts, tz=tz)
    except OSError:
        k = int(-ts // SECONDS + 1)  # Cast avoids "unsupported type for timedelta days: numpy.int32"
        dt = datetime.datetime.fromtimestamp(ts + k * SECONDS, tz=tz) - \
             timedelta(days=k * DAYS)
    return dt



from collections import OrderedDict

from Orange.widgets import widget, gui, settings
from orangecontrib.timeseries import Timeseries, VAR
from orangecontrib.timeseries.widgets._owmodel import OWBaseModel


class OWVARModel(OWBaseModel):
    name = 'VAR Model'
    description = 'Model the time series using vector auto-regression (VAR).'
    icon = 'icons/VAR.svg'
    priority = 220

    maxlags = settings.Setting(1)
    ic = settings.Setting(0)
    trend = settings.Setting(0)

    UserAdviceMessages = [
        widget.Message('Note, VAR model requires the series to be stationary.',
                       'stationary',
                       widget.Message.Warning)
    ]

    IC_LABELS = OrderedDict((('None', None),
                             ("Akaike's information criterion (AIC)", 'aic'),
                             ('Bayesian information criterion (BIC)', 'bic'),
                             ('Hannan–Quinn', 'hqic'),
                             ("Final prediction error (FPE)", 'fpe'),
                             ('Average of the above', 'magic')))
    TREND_LABELS = OrderedDict((('None', 'nc'),
                                ('Constant', 'c'),
                                ('Constant and linear', 'ct'),
                                ('Constant, linear and quadratic', 'ctt')))

    def add_main_layout(self):
        box = gui.vBox(self.controlArea, box='Parameters')
        gui.spin(box, self, 'maxlags', 1, 100, label='Maximum auto-regression order:',
                 callback=self.apply)
        gui.radioButtons(
            box, self, 'ic',
            btnLabels=tuple(self.IC_LABELS.keys()),
            box='Information criterion', label='Optimize AR order by:',
            callback=self.apply)
        gui.radioButtons(
            box, self, 'trend',
            btnLabels=tuple(self.TREND_LABELS.keys()),
            box='Trend', label='Add trend vector(s):',
            callback=self.apply)

    def create_learner(self):
        ic = self.IC_LABELS[tuple(self.IC_LABELS.keys())[self.ic]]
        trend = self.TREND_LABELS[tuple(self.TREND_LABELS.keys())[self.trend]]
        return VAR(self.maxlags, ic, trend)


if __name__ == "__main__":
    from AnyQt.QtWidgets import QApplication

    a = QApplication([])
    ow = OWVARModel()

    data = Timeseries.from_file('airpassengers')
    ow.set_data(data)

    ow.show()
    a.exec()



import unittest

from Orange.data import Table
from Orange.widgets.tests.base import WidgetTest

from orangecontrib.timeseries import Timeseries, ARIMA
from orangecontrib.timeseries.widgets.owmodelevaluation import OWModelEvaluation


class TestOWModelEvaluation(WidgetTest):
    def setUp(self):
        self.widget = self.create_widget(OWModelEvaluation)  # type: OWModelEvaluation

    def test_bad_model(self):
        """
        Do not fail on a bad model.
        GH-37
        """
        w = self.widget
        table = Table("housing")
        time_series = Timeseries.from_data_table(table)
        model = ARIMA((2, 5, 1), 0)
        self.assertFalse(w.Warning.model_not_appropriate.is_shown())
        self.send_signal(w.Inputs.time_series, time_series)
        self.send_signal(w.Inputs.time_series_model, model, 0)
        w.controls.autocommit.click()
        self.assertTrue(w.Warning.model_not_appropriate.is_shown())
        self.send_signal(w.Inputs.time_series_model, None, 0)
        self.assertFalse(w.Warning.model_not_appropriate.is_shown())


if __name__ == "__main__":
    unittest.main()


import unittest
import numpy as np

from orangecontrib.timeseries import Timeseries, seasonal_decompose


data = Timeseries.from_file('airpassengers')


class TestSeasonalDecompose(unittest.TestCase):
    def test_seasonal_decomposition(self):
        decomp = seasonal_decompose(data[:, data.domain.class_var], model='multiplicative')
        self.assertEqual(len(decomp.domain), 4)

        np.testing.assert_almost_equal(
            decomp[:, 'Air passengers (season. adj.)'],
            decomp[:, 'Air passengers (trend)'].X * decomp[:, 'Air passengers (residual)'].X
        )
        np.testing.assert_almost_equal(
            data[:, data.domain.class_var],
            decomp[:, 'Air passengers (season. adj.)'] * decomp[:, 'Air passengers (seasonal)'].X
        )



import unittest
import numpy as np

from Orange.widgets.utils.itemmodels import select_rows
from orangewidget.tests.base import WidgetTest
from Orange.data import Table

from orangecontrib.timeseries import Timeseries
from orangecontrib.timeseries.widgets.owdifference import OWDifference


class TestOWDifference(WidgetTest):

    def setUp(self):
        self.widget = self.create_widget(OWDifference)  # type: OWDifference
        url = "https://datasets.biolab.si/core/ewba-slovenia-illegal-dumpsites.tab"
        self.data = Timeseries.from_url(url)
        self.simple_data = Table('iris')
        self.selected = [self.simple_data.domain.attributes[-1]]

    def test_saved_selection(self):
        self.widget.chosen_operation = self.widget.Operation.DIFF
        self.send_signal(self.widget.Inputs.time_series, self.data)
        output = self.get_output(self.widget.Outputs.time_series)
        # first item is selected and computed
        self.assertEqual(len(output.domain), len(self.data.domain) + 1)
        self.assertTrue("Latitude [°] (diff; order=1)" in output.domain)
        # select some items
        select_rows(self.widget.view, [2, 3, 4])
        output = self.get_output(self.widget.Outputs.time_series)
        self.assertEqual(len(output.domain), len(self.data.domain) + 3)
        # send None
        self.send_signal(self.widget.Inputs.time_series, None)
        self.assertIsNone(self.get_output(self.widget.Outputs.time_series))
        # test if settings are properly restored
        self.send_signal(self.widget.Inputs.time_series, self.data)
        output = self.get_output(self.widget.Outputs.time_series)
        self.assertEqual(len(output.domain), len(self.data.domain) + 3)
        self.assertEqual(len(self.widget.selected), 3)
        self.assertFalse("Latitude [°] (diff; order=1)" in output.domain)

    def test_difference(self):
        w = self.widget
        w.chosen_operation = w.Operation.DIFF
        w.diff_order = 1
        w.shift_period = 1
        w.invert_direction = False
        self.send_signal(w.Inputs.time_series, self.simple_data[:6])
        w.selected = self.selected
        w.commit()
        out = self.get_output(w.Outputs.time_series).X[:, -1]
        true_out = np.asarray([np.nan, 0, 0, 0, 0, 0.2])
        np.testing.assert_array_equal(out, true_out)

        # test order always one, if shift > 1
        w.shift_period = 2
        w.diff_order = 5
        self.send_signal(w.Inputs.time_series, self.simple_data[:6])
        w.selected = self.selected
        w.commit()
        out = self.get_output(w.Outputs.time_series).X[:, -1]
        true_out = np.asarray([np.nan, np.nan, 0, 0, 0, 0.2])
        np.testing.assert_array_equal(out, true_out)

    def test_quotient(self):
        w = self.widget
        w.invert_direction = False
        w.shift_period = 1
        self.send_signal(w.Inputs.time_series, self.simple_data[:6])
        w.chosen_operation = w.Operation.QUOT
        w.selected = self.selected
        w.commit()
        out = self.get_output(w.Outputs.time_series).X[:, -1]
        true_out = np.asarray([np.nan, 1, 1, 1, 1, 2])
        np.testing.assert_array_equal(out, true_out)

    def test_percent(self):
        w = self.widget
        w.invert_direction = False
        w.shift_period = 1
        self.send_signal(w.Inputs.time_series, self.simple_data[:6])
        w.chosen_operation = w.Operation.PERC
        w.selected = self.selected
        w.commit()
        out = self.get_output(w.Outputs.time_series).X[:, -1]
        true_out = np.asarray([np.nan, 0, 0, 0, 0, 100])
        np.testing.assert_array_equal(out, true_out)

    def test_order_spin(self):
        w = self.widget
        w.chosen_operation = w.Operation.DIFF
        w.shift_period = 1
        w.on_changed()
        self.assertTrue(w.order_spin.isEnabled())

        w.shift_period = 2
        w.on_changed()
        self.assertFalse(w.order_spin.isEnabled())

        w.shift_period = 1
        w.on_changed()
        self.assertTrue(w.order_spin.isEnabled())

        w.chosen_operation = w.Operation.QUOT
        w.on_changed()
        self.assertFalse(w.order_spin.isEnabled())

        w.chosen_operation = w.Operation.DIFF
        w.on_changed()
        self.assertTrue(w.order_spin.isEnabled())

        w.chosen_operation = w.Operation.PERC
        w.on_changed()
        self.assertFalse(w.order_spin.isEnabled())


if __name__ == "__main__":
    unittest.main()



import unittest
import numpy as np

from orangecontrib.timeseries import Timeseries, ARIMA, VAR, model_evaluation


data = Timeseries.from_file('airpassengers')


class TestARIMA(unittest.TestCase):
    def test_predict(self):
        model = ARIMA((2, 1, 0))
        model.fit(data)
        forecast, ci95_low, ci95_high = model.predict(10)
        self.assertTrue(
            np.logical_and(forecast > ci95_low, forecast < ci95_high).all())

    def test_predict_as_table(self):
        model = ARIMA((2, 1, 0))
        model.fit(data)
        forecast = model.predict(10, as_table=True)
        self.assertEqual(len(forecast.domain), 1 + 2)


class TestVAR(unittest.TestCase):
    def test_predict(self):
        model = VAR(2)
        model.fit(data)
        forecast, ci95_low, ci95_high = model.predict(10)
        self.assertTrue(
            np.logical_and(forecast > ci95_low, forecast < ci95_high).all())

    def test_predict_as_table(self):
        model = VAR(2)
        model.fit(data)
        forecast = model.predict(10, as_table=True)
        self.assertEqual(len(forecast.domain), 2 * (1 + 2))


class TestModelEvaluation(unittest.TestCase):
    def test_model_evaluation(self):
        models = [ARIMA((1, 1, 0)), ARIMA((2, 1, 1)), VAR(1), VAR(3)]
        results = model_evaluation(data, models, n_folds=10, forecast_steps=3)
        results = np.array(results, dtype=object)
        self.assertEqual(results.shape, (4 * 2 + 1, 8))
        np.testing.assert_equal(results[1:, 1:].astype(float) > 0, True)



import unittest

from Orange.data import Table
from Orange.widgets.tests.base import WidgetTest

from orangecontrib.timeseries import Timeseries
from orangecontrib.timeseries.widgets.owspiralogram import OWSpiralogram
from Orange.widgets.tests.utils import simulate
from Orange.widgets.settings import Context


class TestOWSpiralogram(WidgetTest):
    def setUp(self):
        self.widget = self.create_widget(OWSpiralogram)  # type: OWSpiralogram
        self.passengers = Timeseries.from_file("airpassengers")
        self.philadelphia = Timeseries.from_url('http://datasets.orange.biolab.si/core/philadelphia-crime.csv.xz')

    @staticmethod
    def select_item(widget, index):
        widget.indices = [index]
        widget.commit()

    def test_new_data(self):
        """
        Widget crashes when it gets new data with different domain.
        GH-50
        """
        w = self.widget
        url = "http://file.biolab.si/datasets/cyber-security-breaches.tab"
        time_series2 = Timeseries.from_url(url)
        self.send_signal(w.Inputs.time_series, self.passengers)
        self.select_item(w, 0)
        output1 = self.get_output(w.Outputs.time_series)
        self.send_signal(w.Inputs.time_series, time_series2)
        self.select_item(w, 0)
        output2 = self.get_output(w.Outputs.time_series)
        self.assertNotEqual(output1, output2)

    def test_no_datetime(self):
        """ Raise error if no data with TimeVariable """
        w = self.widget
        table = Table.from_file('iris')
        self.send_signal(w.Inputs.time_series, self.passengers)
        self.assertFalse(w.Error.no_time_variable.is_shown())
        self.send_signal(w.Inputs.time_series, table)
        self.assertTrue(w.Error.no_time_variable.is_shown())
        self.send_signal(w.Inputs.time_series, self.passengers)
        self.assertFalse(w.Error.no_time_variable.is_shown())

    def test_tz_aggregation(self):
        """ Aggregation should consider timezone """
        w = self.widget
        data = self.passengers[:20]
        self.send_signal(w.Inputs.time_series, data)
        # select the first item
        self.select_item(w, 0)
        output = self.get_output(w.Outputs.time_series)
        self.assertEqual(output[0][0], "1949-01-01")

    def test_cb_axes(self):
        """ Test that all possible axes work, including discrete variables. """
        w = self.widget
        self.send_signal(w.Inputs.time_series, self.philadelphia)
        # test all possibilities for Y axis
        simulate.combobox_run_through_all(w.combo_ax2)
        # test all possibilities for radial
        simulate.combobox_run_through_all(w.combo_ax1)

    def test_time_variable(self):
        """ Spiralogram should work with TimeVariable. """
        w = self.widget
        self.send_signal(w.Inputs.time_series, self.philadelphia)
        # select time variable
        simulate.combobox_activate_item(w.attr_cb, 'Datetime')
        # test all possibilities for aggregations
        simulate.combobox_run_through_all(w.combo_func)

    def test_migrate_settings_from_version_1_disc(self):
        settings = {
            '__version__': 1,
            'context_settings': [
                Context(attributes={'Type': 1, "foo": 2},
                        metas={'name': 4},
                        values={'agg_attr': ([('Type', 101)], -3),
                                'agg_func': (0, -2),
                                'ax1': ('months', -2),
                                'ax2': ('years', -2)})],
            'controlAreaVisible': True,
            'invert_date_order': False,
            'savedWidgetGeometry': None
        }
        w = self.create_widget(OWSpiralogram, stored_settings=settings)
        self.send_signal(w.Inputs.time_series, self.philadelphia, widget=w)
        self.assertEqual(w.agg_attr.name, 'Type')
        self.assertEqual(w.agg_func, 'Mode')

    def test_migrate_settings_from_version_1_time(self):
        settings = {
            '__version__': 1,
            'context_settings': [
                Context(
                    attributes={'Datetime': 1, "foo": 2},
                    metas={'name': 4},
                    values={'agg_attr': ([('Datetime', 104)], -3),
                            'agg_func': (12, -2),
                            'ax1': ('months', -2),
                            'ax2': ('years', -2)})
            ],
            'controlAreaVisible': True,
            'invert_date_order': False,
            'savedWidgetGeometry': None
        }
        w = self.create_widget(OWSpiralogram, stored_settings=settings)
        self.send_signal(w.Inputs.time_series, self.philadelphia, widget=w)
        self.assertEqual(w.agg_attr.name, 'Datetime')
        self.assertEqual(w.agg_func, 'Mean')


if __name__ == "__main__":
    unittest.main()



import calendar
import logging
from datetime import datetime, timedelta
from numbers import Number

log = logging.getLogger(__name__)


def cache_clears(*caches):
    def decorator(func):
        def f(*args, **kwargs):
            for cache in caches:
                # Consider property getters also
                getattr(cache, 'fget', cache).cache_clear()
            return func(*args, **kwargs)
        return f
    return decorator


def add_time(start_dt: datetime.date, delta, quantity):
    if isinstance(delta, Number):
        return start_dt + timedelta(seconds=delta*quantity)
    quantity = delta[0] * quantity
    if delta:
        if delta[1] == 'month':
            years = int(quantity / 12)
            months = quantity - years * 12
            months_result = start_dt.month + months
            if months_result < 1:
                years -= 1
                months_result += 12
            elif 12 < months_result:
                years += 1
                months_result -= 12
            years_result = start_dt.year + years
            last_calendar_day = calendar.monthrange(years_result, months_result)[1]
            return start_dt.replace(
                day=min(start_dt.day, last_calendar_day),
                month=months_result,
                year=years_result
            )
        else:  # elif delta[1] == 'year':
            return start_dt.replace(
                year=start_dt.year + quantity,
            )
    log.warning('"None" timedelta supplied when adding time, '
                'not adding any time')
    return start_dt


from AnyQt.QtCore import Qt

from Orange.data import Table
from Orange.widgets import widget, gui, settings
from Orange.widgets.settings import Setting
from Orange.widgets.utils.itemmodels import PyTableModel
from Orange.widgets.widget import Input, Output, OWWidget, Msg, AttributeList
from Orange.widgets.utils.concurrent import TaskState, ConcurrentWidgetMixin

from orangecontrib.timeseries import Timeseries, granger_causality
from orangewidget.utils.widgetpreview import WidgetPreview


COLUMNS = ["Min. lag", "p-value", "Series 1", "", "Series 2"]


def run(data: Table, max_lag: int, confidence: int, state: TaskState):
    def advance(progress: float):
        if state.is_interruption_requested():
            raise Exception
        state.set_progress_value(progress * 100)

    res = granger_causality(
        data, max_lag, 1 - confidence / 100, callback=advance
    )
    return [[lag, pval, row, "→", col] for lag, pval, row, col in res]


class OWGrangerCausality(OWWidget, ConcurrentWidgetMixin):
    name = "Granger Causality"
    description = (
        "Test if one time series Granger-causes (i.e. can be an "
        "indicator of) another."
    )
    icon = "icons/GrangerCausality.svg"
    priority = 190

    class Inputs:
        time_series = Input("Time series", Table, replaces=["Timeseries"])

    class Outputs:
        selected_features = Output("Selected features", AttributeList)

    class Information(OWWidget.Information):
        modified = Msg(
            "The parameter settings have been changed. Press "
            '"Test" to rerun with the new settings.'
        )

    max_lag = Setting(20)
    confidence = Setting(95)
    autocommit = Setting(True)
    sorting = Setting((1, Qt.AscendingOrder))

    UserAdviceMessages = [
        widget.Message(
            "We say <i>X</i> Granger-causes <i>Y</i> if "
            "predictions of values of <i>Y</i> based on its own "
            "past values and on the past values of <i>X</i> are "
            "better than predictions of <i>Y</i> based on its "
            "past values alone.<br><br>"
            "It does NOT mean <i>X</i> causes <i>Y</i>!",
            "explanation",
            widget.Message.Warning,
        )
    ]

    class Error(widget.OWWidget.Error):
        unexpected_error = widget.Msg("Unexpected error: {}")

    def __init__(self):
        OWWidget.__init__(self)
        ConcurrentWidgetMixin.__init__(self)
        self.data = None
        self.selected_attributes = None
        box = gui.vBox(self.controlArea, "Granger Test")
        gui.hSlider(
            box,
            self,
            "confidence",
            minValue=90,
            maxValue=99,
            label="Confidence:",
            labelFormat=" %d%%",
            callback=self._setting_changed,
        )
        gui.spin(
            box,
            self,
            "max_lag",
            1,
            50,
            label="Max lag:",
            callback=self._setting_changed,
        )
        self.test_button = gui.button(box, self, "&Test", self._toggle_run)
        gui.rubber(self.controlArea)

        self.model = model = PyTableModel(parent=self)

        model.setHorizontalHeaderLabels(COLUMNS)
        self.causality_view = view = gui.TableView(self)
        view.setModel(model)
        bold = view.BoldFontDelegate(self)
        view.setItemDelegateForColumn(2, bold)
        view.setItemDelegateForColumn(4, bold)
        view.horizontalHeader().setStretchLastSection(False)
        view.horizontalHeader().sectionClicked.connect(self.header_click)
        view.selectionModel().selectionChanged.connect(self.on_select)
        view.sortByColumn(1, Qt.AscendingOrder)
        self.mainArea.layout().addWidget(view)
        self._set_modified(False)

        self.auto_commit_widget = gui.auto_commit(
            widget=self.controlArea,
            master=self,
            value="autocommit",
            label="Apply",
            commit=self.commit,
        )
        # TODO: allow setting filters or choosing what variables to include in test

    def _set_modified(self, state):
        self.Information.modified(shown=state)

    def _toggle_run(self):
        if self.task is not None:
            self._invalidate_run()
        else:
            self._run()

    def _setting_changed(self):
        self._set_modified(True)
        self._invalidate_run()

    def _invalidate_run(self):
        self.cancel()
        self.test_button.setText("Test")

    def _run(self):
        self._set_modified(False)
        self.model.clear()
        self.selected_attributes = None
        self.Error.unexpected_error.clear()
        if self.data is None:
            return
        self.start(run, self.data, self.max_lag, self.confidence)
        self.test_button.setText("Stop")

    def on_done(self, res):
        self.model.wrap(res)
        self.test_button.setText("Test")

        # Re-apply sort
        try:
            sort_column, sort_order = self.sorting
            # adds 1 for '#' (discrete count) column
            self.model.sort(sort_column, sort_order)
            self.causality_view.horizontalHeader().setSortIndicator(
                sort_column, sort_order
            )
        except ValueError:
            pass

    def on_exception(self, ex):
        self.Error.unexpected_error(ex.args[0])
        self.test_button.setText("Test")

    @Inputs.time_series
    def set_data(self, data):
        self.data = None if data is None else Timeseries.from_data_table(data)
        self.selected_attributes = None
        self._run()
        self.Outputs.selected_features.send(None)

    def commit(self):
        self.Outputs.selected_features.send(self.selected_attributes or None)

    def on_select(self):
        selected_rows = self.causality_view.selectionModel().selectedRows()
        row_indices = [i.row() for i in selected_rows]
        row_indices = self.model.mapToSourceRows(row_indices)
        attributes = [
            self.model[i][j]
            for i in row_indices
            for j in [COLUMNS.index("Series 1"), COLUMNS.index("Series 2")]
        ]
        # remove duplicated attributes - I know I could use set for this but
        # I want to keep the order of attributes - keeping first appearance
        # of an attribute and remove the rest
        attributes = list(dict.fromkeys(attributes))
        self.selected_attributes = [self.data.domain[a] for a in attributes]
        self.commit()

    def header_click(self, _):
        # Store the header states
        sort_order = self.model.sortOrder()
        sort_column = self.model.sortColumn()
        self.sorting = (sort_column, sort_order)


if __name__ == "__main__":
    data = Timeseries.from_file("AMZN")
    WidgetPreview(OWGrangerCausality).run(data)



import datetime
from contextlib import contextmanager

from collections import OrderedDict
from numbers import Number
from os.path import join, dirname

from AnyQt.QtWidgets import QLabel, QDateTimeEdit
from AnyQt.QtCore import QDateTime, Qt, QSize, QTimer, QTimeZone
from AnyQt.QtGui import QFontDatabase, QFont

from Orange.data import Table, TimeVariable
from Orange.widgets import widget, gui, settings
from Orange.widgets.widget import Input, Output

from orangecontrib.timeseries import Timeseries
from orangecontrib.timeseries.util import add_time
from orangecontrib.timeseries.widgets._rangeslider import ViolinSlider


class _TimeSliderMixin:
    DEFAULT_SCALE_LENGTH = 500

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.__scale_minimum = None  # type: datetime.date
        self.__scale_maximum = None  # type: datetime.date
        self.__time_delta = None  # type: Optional[Union[Number, tuple]]
        self.__formatter = str  # type: Callable

    def setScale(self, minimum, maximum, time_delta):
        self.__scale_minimum = datetime.datetime.fromtimestamp(
            round(minimum),
            tz=datetime.timezone.utc
        )
        self.__scale_maximum = datetime.datetime.fromtimestamp(
            round(maximum),
            tz=datetime.timezone.utc
        )
        self.__time_delta = time_delta

    def scale(self, value):
        quantity = round(value - self.minimum())
        delta = self.__time_delta
        if delta is None:
            return (self.__scale_minimum +
                    (self.__scale_maximum - self.__scale_minimum) *
                    (value - self.minimum()) /
                    (self.maximum() - self.minimum())).timestamp()
        scaled_dt = add_time(self.__scale_minimum, delta, quantity)
        return scaled_dt.timestamp()

    def unscale(self, value):
        # Unscale e.g. absolute time to slider value
        dt_val = datetime.datetime.fromtimestamp(round(value), tz=datetime.timezone.utc)
        delta = self.__time_delta
        if isinstance(delta, Number):
            diff = dt_val - self.__scale_minimum
            return round(diff / datetime.timedelta(milliseconds=1000 * delta))
        elif delta:
            if delta[1] == 'month':
                return self.minimum() + delta[0] \
                       * ((dt_val.year - self.__scale_minimum.year) * 12
                          + max(dt_val.month - self.__scale_minimum.month,
                                0))
            else:  # elif delta[1] == 'year':
                return self.minimum() + delta[0] \
                       * (dt_val.year - self.__scale_minimum.year)

        return self.minimum() + ((dt_val - self.__scale_minimum) *
                                 (self.maximum() - self.minimum()) /
                                 (self.__scale_maximum - self.__scale_minimum) + self.minimum())

    def setFormatter(self, formatter):
        self.__formatter = formatter

    def formatValues(self, minValue, maxValue):
        if None in [self.__scale_minimum, self.__scale_maximum]:
            return '0', '0'
        return (self.__formatter(int(self.scale(minValue))),
                self.__formatter(int(self.scale(maxValue))))


class Slider(_TimeSliderMixin, ViolinSlider):
    def sizeHint(self):
        return QSize(200, 100)


@contextmanager
def blockSignals(*objects):
    for obj in objects:
        obj.blockSignals(True)
    yield
    for obj in objects:
        obj.blockSignals(False)


def load_icons_font():
    fontId = QFontDatabase.addApplicationFont(
        join(dirname(__file__), 'icons', 'TimeSlice-icons.ttf')
    )
    font_family = QFontDatabase.applicationFontFamilies(fontId)[0]
    return QFont(font_family, 9, QFont.Normal)


class OWTimeSlice(widget.OWWidget):
    name = 'Time Slice'
    description = 'Select a slice of measurements on a time interval.'
    icon = 'icons/TimeSlice.svg'
    priority = 550

    class Inputs:
        data = Input("Data", Table)

    class Outputs:
        subset = Output("Subset", Table)

    settings_version = 2

    want_main_area = False

    class Error(widget.OWWidget.Error):
        no_time_variable = widget.Msg('Data contains no time variable')
        no_time_delta = widget.Msg('Data contains only one time point')

    MAX_SLIDER_VALUE = 500
    DATE_FORMATS = ('yyyy', '-MM', '-dd', '  HH:mm:ss.zzz')
    # only appropriate overlap amounts are shown, but these are all the options
    DELAY_VALUES = (0.1, 0.2, 0.5, 1, 2, 5, 10, 15, 30)
    STEP_SIZES = OrderedDict((
        ('1 second', 1),
        ('5 seconds', 5),
        ('10 seconds', 10),
        ('15 seconds', 15),
        ('30 seconds', 30),
        ('1 minute', 60),
        ('5 minutes', 300),
        ('10 minutes', 600),
        ('15 minutes', 900),
        ('30 minutes', 1800),
        ('1 hour', 3600),
        ('2 hours', 7200),
        ('3 hours', 10800),
        ('6 hours', 21600),
        ('12 hours', 43200),
        ('1 day', 86400),
        ('1 week', 604800),
        ('2 weeks', 1209600),
        ('1 month', (1, 'month')),
        ('2 months', (2, 'month')),
        ('3 months', (3, 'month')),
        ('6 months', (6, 'month')),
        ('1 year', (1, 'year')),
        ('2 years', (2, 'year')),
        ('5 years', (5, 'year')),
        ('10 years', (10, 'year')),
        ('25 years', (25, 'year')),
        ('50 years', (50, 'year')),
        ('100 years', (100, 'year'))))

    loop_playback = settings.Setting(True)
    custom_step_size = settings.Setting(False)
    step_size = settings.Setting(next(iter(STEP_SIZES)))
    playback_interval = settings.Setting(1)
    slider_values = settings.Setting((0, .2 * MAX_SLIDER_VALUE))

    icons_font = None

    def __init__(self):
        super().__init__()
        self._delta = 0
        self.play_timer = QTimer(self,
                                 interval=1000*self.playback_interval,
                                 timeout=self.play_single_step)
        slider = self.slider = Slider(Qt.Horizontal, self,
                                      minimum=0, maximum=self.MAX_SLIDER_VALUE,
                                      tracking=True,
                                      playbackInterval=1000*self.playback_interval,
                                      valuesChanged=self.sliderValuesChanged,
                                      minimumValue=self.slider_values[0],
                                      maximumValue=self.slider_values[1])
        slider.setShowText(False)
        selectBox = gui.vBox(self.controlArea, 'Select a Time Range')
        selectBox.layout().addWidget(slider)

        dtBox = gui.hBox(selectBox)

        kwargs = dict(calendarPopup=True,
                      displayFormat=' '.join(self.DATE_FORMATS),
                      timeSpec=Qt.UTC)
        date_from = self.date_from = QDateTimeEdit(self, **kwargs)
        date_to = self.date_to = QDateTimeEdit(self, **kwargs)

        def datetime_edited(dt_edit):
            minTime = self.date_from.dateTime().toMSecsSinceEpoch() / 1000
            maxTime = self.date_to.dateTime().toMSecsSinceEpoch() / 1000
            if minTime > maxTime:
                minTime = maxTime = minTime if dt_edit == self.date_from else maxTime
                other = self.date_to if dt_edit == self.date_from else self.date_from
                with blockSignals(other):
                    other.setDateTime(dt_edit.dateTime())

            self.dteditValuesChanged(minTime, maxTime)

        date_from.dateTimeChanged.connect(lambda: datetime_edited(date_from))
        date_to.dateTimeChanged.connect(lambda: datetime_edited(date_to))

        # hotfix, does not repaint on click of arrow
        date_from.calendarWidget().currentPageChanged.connect(
            lambda: date_from.calendarWidget().repaint()
        )
        date_to.calendarWidget().currentPageChanged.connect(
            lambda: date_to.calendarWidget().repaint()
        )

        dtBox.layout().addStretch(100)
        dtBox.layout().addWidget(date_from)
        dtBox.layout().addWidget(QLabel(' – '))
        dtBox.layout().addWidget(date_to)
        dtBox.layout().addStretch(100)

        hCenterBox = gui.hBox(self.controlArea)
        gui.rubber(hCenterBox)
        vControlsBox = gui.vBox(hCenterBox)

        stepThroughBox = gui.vBox(vControlsBox, 'Step/Play Through')
        gui.rubber(stepThroughBox)
        gui.checkBox(stepThroughBox, self, 'loop_playback',
                     label='Loop playback')
        customStepBox = gui.hBox(stepThroughBox)
        gui.checkBox(customStepBox, self, 'custom_step_size',
                     label='Custom step size: ',
                     toolTip='If not chosen, the active interval moves forward '
                             '(backward), stepping in increments of its own size.')
        self.stepsize_combobox = gui.comboBox(customStepBox, self, 'step_size',
                                              items=tuple(self.STEP_SIZES.keys()),
                                              sendSelectedValue=True)
        playBox = gui.hBox(stepThroughBox)
        gui.rubber(playBox)
        gui.rubber(stepThroughBox)

        if self.icons_font is None:
            self.icons_font = load_icons_font()

        self.step_backward = gui.button(playBox, self, '⏪',
                                        callback=lambda: self.play_single_step(backward=True),
                                        autoDefault=False)
        self.step_backward.setFont(self.icons_font)
        self.play_button = gui.button(playBox, self, '▶️',
                                      callback=self.playthrough,
                                      toggleButton=True, default=True)
        self.play_button.setFont(self.icons_font)
        self.step_forward = gui.button(playBox, self, '⏩',
                                       callback=self.play_single_step,
                                       autoDefault=False)
        self.step_forward.setFont(self.icons_font)

        gui.rubber(playBox)
        intervalBox = gui.vBox(vControlsBox, 'Playback/Tracking interval')
        intervalBox.setToolTip('In milliseconds, set the delay for playback and '
                               'for sending data upon manually moving the interval.')

        def set_intervals():
            self.play_timer.setInterval(1000 * self.playback_interval)
            self.slider.tracking_timer.setInterval(1000 * self.playback_interval)
        gui.valueSlider(intervalBox, self, 'playback_interval',
                        label='Delay:', labelFormat='%.2g sec',
                        values=self.DELAY_VALUES,
                        callback=set_intervals)

        gui.rubber(hCenterBox)
        gui.rubber(self.controlArea)
        self._set_disabled(True)

    def sliderValuesChanged(self, minValue, maxValue):
        self._delta = max(1, (maxValue - minValue))
        minTime = self.slider.scale(minValue)
        maxTime = self.slider.scale(maxValue)

        from_dt = QDateTime.fromMSecsSinceEpoch(minTime * 1000).toUTC()
        to_dt = QDateTime.fromMSecsSinceEpoch(maxTime * 1000).toUTC()
        if self.date_from.dateTime() != from_dt:
            with blockSignals(self.date_from):
                self.date_from.setDateTime(from_dt)
        if self.date_from.dateTime() != to_dt:
            with blockSignals(self.date_to):
                self.date_to.setDateTime(to_dt)

        self.send_selection(minTime, maxTime)

    def dteditValuesChanged(self, minTime, maxTime):
        minValue = self.slider.unscale(minTime)
        maxValue = self.slider.unscale(maxTime)
        if minValue == maxValue:
            # maxValue's range is minValue's range shifted by one
            maxValue += 1
            maxTime = self.slider.scale(maxValue)
            to_dt = QDateTime.fromMSecsSinceEpoch(maxTime * 1000).toUTC()
            with blockSignals(self.date_to):
                self.date_to.setDateTime(to_dt)

        self._delta = max(1, (maxValue - minValue))

        if self.slider_values != (minValue, maxValue):
            self.slider_values = (minValue, maxValue)
            with blockSignals(self.slider):
                self.slider.setValues(minValue, maxValue)

        self.send_selection(minTime, maxTime)

    def send_selection(self, minTime, maxTime):
        try:
            time_values = self.data.time_values
        except AttributeError:
            return
        indices = (minTime <= time_values) & (time_values < maxTime)
        self.Outputs.subset.send(self.data[indices] if indices.any() else None)

    def playthrough(self):
        playing = self.play_button.isChecked()

        for widget in (self.slider,
                       self.step_forward,
                       self.step_backward):
            widget.setDisabled(playing)

        for widget in (self.date_from,
                       self.date_to):
            widget.setReadOnly(playing)

        if playing:
            self.play_timer.start()
            self.play_button.setText('⏸')
        else:
            self.play_timer.stop()
            self.play_button.setText('▶️')

        # hotfix
        self.repaint()

    def play_single_step(self, backward=False):
        minValue, maxValue = self.slider.values()
        orig_delta = delta = self._delta

        def new_value(value):
            if self.custom_step_size:
                step_amount = self.STEP_SIZES[self.step_size]
                time = datetime.datetime.fromtimestamp(self.slider.scale(value),
                                                       tz=datetime.timezone.utc)
                newTime = add_time(time, step_amount, -1 if backward else 1)
                return self.slider.unscale(newTime.timestamp())
            return value + (-delta if backward else delta)

        if maxValue == self.slider.maximum() and not backward:
            minValue = self.slider.minimum()
            maxValue = self.slider.minimum() + delta

            if not self.loop_playback:
                self.play_button.click()
                assert not self.play_timer.isActive()
                assert not self.play_button.isChecked()

        elif minValue == self.slider.minimum() and backward:
            maxValue = self.slider.maximum()
            minValue = min(self.slider.maximum(), new_value(maxValue))
        else:
            minValue = min(new_value(minValue), self.slider.maximum())
            maxValue = min(new_value(maxValue), self.slider.maximum())
        # Blocking signals because we want this to be synchronous to avoid
        # re-setting self._delta
        with blockSignals(self.slider):
            self.slider.setValues(minValue, maxValue)
        self.sliderValuesChanged(self.slider.minimumValue(), self.slider.maximumValue())
        self._delta = orig_delta  # Override valuesChanged handler

        # hotfix
        self.slider.repaint()

    def _set_disabled(self, is_disabled):
        if is_disabled and self.play_timer.isActive():
            self.play_button.click()
            assert not self.play_timer.isActive()
            assert not self.play_button.isChecked()

        for func in [self.date_from, self.date_to, self.step_backward,
                     self.play_button, self.step_forward,
                     self.controls.loop_playback, self.controls.step_size,
                     self.controls.playback_interval, self.slider]:
            func.setDisabled(is_disabled)

    @Inputs.data
    def set_data(self, data):
        slider = self.slider
        self.data = data = None if data is None else Timeseries.from_data_table(data)

        def disabled():
            slider.setFormatter(str)
            slider.setHistogram(None)
            slider.setScale(0, 0, None)
            slider.setValues(0, 0)
            self._set_disabled(True)
            self.Outputs.subset.send(None)

        if data is None:
            disabled()
            return

        if not isinstance(data.time_variable, TimeVariable):
            self.Error.no_time_variable()
            disabled()
            return
        if not data.time_delta.deltas:
            self.Error.no_time_delta()
            disabled()
            return
        self.Error.clear()
        var = data.time_variable

        time_values = data.time_values

        min_dt = datetime.datetime.fromtimestamp(round(time_values.min()), tz=datetime.timezone.utc)
        max_dt = datetime.datetime.fromtimestamp(round(time_values.max()), tz=datetime.timezone.utc)

        # Depending on time delta:
        #   - set slider maximum (granularity)
        #   - set range for end dt (+ 1 timedelta)
        #   - set date format
        #   - set time overlap options
        delta = data.time_delta.gcd
        range = max_dt - min_dt
        if isinstance(delta, Number):
            maximum = round(range.total_seconds() / delta)

            timedelta = datetime.timedelta(milliseconds=delta * 1000)
            min_dt2 = min_dt + timedelta
            max_dt2 = max_dt + timedelta

            if delta >= 86400:  # more than a day
                date_format = ''.join(self.DATE_FORMATS[0:3])
            else:
                date_format = ''.join(self.DATE_FORMATS)

            for k, n in [(k, n) for k, n in self.STEP_SIZES.items()
                         if isinstance(n, Number)]:
                if delta <= n:
                    min_overlap = k
                    break
            else:
                min_overlap = '1 day'
        else:  # isinstance(delta, tuple)
            if delta[1] == 'month':
                months = (max_dt.year - min_dt.year) * 12 + \
                         (max_dt.month - min_dt.month)
                maximum = months / delta[0]

                if min_dt.month < 12 - delta[0]:
                    min_dt2 = min_dt.replace(
                        month=min_dt.month + delta[0]
                    )
                else:
                    min_dt2 = min_dt.replace(
                        year=min_dt.year + 1,
                        month=12 - min_dt.month + delta[0]
                    )
                if max_dt.month < 12 - delta[0]:
                    max_dt2 = max_dt.replace(
                        month=max_dt.month + delta[0]
                    )
                else:
                    max_dt2 = max_dt.replace(
                        year=max_dt.year + 1,
                        month=12 - min_dt.month + delta[0]
                    )

                date_format = ''.join(self.DATE_FORMATS[0:2])

                for k, (i, u) in [(k, v) for k, v in self.STEP_SIZES.items()
                                  if isinstance(v, tuple) and v[1] == 'month']:
                    if delta[0] <= i:
                        min_overlap = k
                        break
                else:
                    min_overlap = '1 year'
            else:  # elif delta[1] == 'year':
                years = max_dt.year - min_dt.year
                maximum = years / delta[0]

                min_dt2 = min_dt.replace(
                    year=min_dt.year + delta[0],
                )
                max_dt2 = max_dt.replace(
                    year=max_dt.year + delta[0],
                )

                date_format = self.DATE_FORMATS[0]

                for k, (i, u) in [(k, v) for k, v in self.STEP_SIZES.items()
                                  if isinstance(v, tuple) and v[1] == 'year']:
                    if delta[0] <= i:
                        min_overlap = k
                        break
                else:
                    raise Exception('Timedelta larger than 100 years')

        # find max sensible time overlap
        upper_overlap_limit = range / 2
        for k, overlap in self.STEP_SIZES.items():
            if isinstance(overlap, Number):
                if upper_overlap_limit.total_seconds() <= overlap:
                    max_overlap = k
                    break
            else:
                i, u = overlap
                if u == 'month':
                    month_diff = (max_dt.year - min_dt.year) * 12 \
                                 + max(0, max_dt.month - min_dt.month)
                    if month_diff / 2 <= i:
                        max_overlap = k
                        break
                else:  # if u == 'year':
                    year_diff = max_dt.year - min_dt.year
                    if year_diff / 2 <= i:
                        max_overlap = k
                        break
        else:
            # last item in step sizes
            *_, max_overlap = self.STEP_SIZES.keys()

        self.stepsize_combobox.clear()
        dict_iter = iter(self.STEP_SIZES.keys())
        next_item = next(dict_iter)
        while next_item != min_overlap:
            next_item = next(dict_iter)
        self.stepsize_combobox.addItem(next_item)
        self.step_size = next_item
        while next_item != max_overlap:
            next_item = next(dict_iter)
            self.stepsize_combobox.addItem(next_item)

        slider.setMinimum(0)
        slider.setMaximum(maximum + 1)

        self._set_disabled(False)
        slider.setHistogram(time_values)
        slider.setFormatter(var.repr_val)
        slider.setScale(time_values.min(), time_values.max(), data.time_delta.gcd)
        self.sliderValuesChanged(slider.minimumValue(), slider.maximumValue())

        def utc_dt(dt):
            qdt = QDateTime(dt)
            qdt.setTimeZone(QTimeZone.utc())
            return qdt

        self.date_from.setDateTimeRange(utc_dt(min_dt), utc_dt(max_dt))
        self.date_to.setDateTimeRange(utc_dt(min_dt2), utc_dt(max_dt2))
        self.date_from.setDisplayFormat(date_format)
        self.date_to.setDisplayFormat(date_format)

        def format_time(i):
            dt = QDateTime.fromMSecsSinceEpoch(i * 1000).toUTC()
            return dt.toString(date_format)

        self.slider.setFormatter(format_time)

    @classmethod
    def migrate_settings(cls, settings_, version):
        if version < 2:
            interval = settings_["playback_interval"] / 1000
            if interval in cls.DELAY_VALUES:
                settings_["playback_interval"] = interval
            else:
                settings_["playback_interval"] = 1


if __name__ == '__main__':
    from AnyQt.QtWidgets import QApplication
    app = QApplication([])
    w = OWTimeSlice()
    data = Table('airpassengers')
    w.set_data(data)
    w.show()
    app.exec()



from datetime import datetime, timedelta, date

from AnyQt.QtCore import QDate
from AnyQt.QtWidgets import QDateEdit, QComboBox

from Orange.widgets import widget, gui, settings
from Orange.widgets.widget import Output

from orangecontrib.timeseries import Timeseries
from orangecontrib.timeseries.datasources import finance_data


class OWYahooFinance(widget.OWWidget):
    name = 'Yahoo Finance'
    description = "Generate time series from Yahoo Finance stock market data."
    icon = 'icons/YahooFinance.svg'
    priority = 9

    class Outputs:
        time_series = Output("Time series", Timeseries)

    QT_DATE_FORMAT = 'yyyy-MM-dd'
    PY_DATE_FORMAT = '%Y-%m-%d'
    MIN_DATE = date(1851, 1, 1)

    date_from = settings.Setting(
        (datetime.now().date() - timedelta(5 * 365)).strftime(PY_DATE_FORMAT))
    date_to = settings.Setting(datetime.now().date().strftime(PY_DATE_FORMAT))
    symbols = settings.Setting(['AMZN', 'AAPL', 'GOOG', 'FB', 'SPY', '^DJI', '^TNX'])

    want_main_area = False
    resizing_enabled = False

    class Error(widget.OWWidget.Error):
        download_error = widget.Msg('Failed to download data (HTTP Error {}). '
                                    'Wrong stock symbol?')

    def __init__(self):
        box = gui.widgetBox(self.controlArea, 'Yahoo Finance Stock Data',
                            orientation='horizontal')
        lbox = gui.widgetBox(box, orientation='vertical')
        hbox = gui.widgetBox(lbox, orientation='horizontal')
        gui.label(hbox, self, 'Ticker:')
        self.combo = combo = QComboBox(editable=True,
                                       insertPolicy=QComboBox.InsertAtTop)
        combo.addItems(self.symbols)
        hbox.layout().addWidget(combo)
        # combo = gui.comboBox(
        #     lbox, self, 'symbol',#, items=self.symbols,
        #     label='Ticker:', orientation='horizontal',
        #     editable=True, maximumContentsLength=-1)
        gui.rubber(combo.parentWidget())
        minDate = QDate.fromString(self.MIN_DATE.strftime(self.PY_DATE_FORMAT),
                                   self.QT_DATE_FORMAT)
        date_from = QDateEdit(
            QDate.fromString(self.date_from, self.QT_DATE_FORMAT),
            displayFormat=self.QT_DATE_FORMAT,
            minimumDate=minDate,
            calendarPopup=True)
        date_to = QDateEdit(
            QDate.fromString(self.date_to, self.QT_DATE_FORMAT),
            displayFormat=self.QT_DATE_FORMAT,
            minimumDate=minDate,
            calendarPopup=True)
        date_from.dateChanged.connect(
            lambda date: setattr(self, 'date_from',
                                 date.toString(self.QT_DATE_FORMAT)))
        date_to.dateChanged.connect(
            lambda date: setattr(self, 'date_to',
                                 date.toString(self.QT_DATE_FORMAT)))
        hbox = gui.hBox(lbox)
        gui.label(hbox, self, "From:")
        hbox.layout().addWidget(date_from)
        hbox = gui.hBox(lbox)
        gui.label(hbox, self, "To:")
        hbox.layout().addWidget(date_to)

        self.button = gui.button(self.controlArea, self, 'Download',
                                 callback=self.download)

    def download(self):
        date_from = datetime.strptime(self.date_from, self.PY_DATE_FORMAT)
        date_to = datetime.strptime(self.date_to, self.PY_DATE_FORMAT)

        # Update symbol in symbols history
        symbol = self.combo.currentText().strip().upper()
        self.combo.removeItem(self.combo.currentIndex())
        self.combo.insertItem(0, symbol)
        self.combo.setCurrentIndex(0)
        try:
            self.symbols.remove(symbol)
        except ValueError:
            pass
        self.symbols.insert(0, symbol)

        if not symbol:
            return

        self.Error.clear()
        with self.progressBar(3) as progress:
            try:
                progress.advance()
                self.button.setDisabled(True)
                data = finance_data(symbol, date_from, date_to)

                self.Outputs.time_series.send(data)
            except Exception as e:
                self.Error.download_error(getattr(e, 'status', -1))
            finally:
                self.button.setDisabled(False)


if __name__ == "__main__":
    from AnyQt.QtWidgets import QApplication

    a = QApplication([])
    ow = OWYahooFinance()

    ow.show()
    a.exec()



from numbers import Number
from collections import OrderedDict
from os.path import join, dirname
from typing import List

import numpy as np
from AnyQt.QtWidgets import (
    QTreeWidget,
    QWidget,
    QPushButton,
    QListView,
    QVBoxLayout,
)
from AnyQt.QtGui import QIcon, QCloseEvent
from AnyQt.QtCore import QSize, pyqtSignal

from orangewidget.utils.widgetpreview import WidgetPreview
from Orange.data import TimeVariable, Table, Variable
from Orange.widgets import gui
from Orange.widgets.widget import OWWidget, Input, AttributeList
from Orange.widgets.utils import vartype
from Orange.widgets.settings import (
    ContextSetting,
    IncompatibleContext,
    DomainContextHandler,
)
from Orange.widgets.utils.itemmodels import VariableListModel

from orangecontrib.timeseries import Timeseries
from orangecontrib.timeseries.widgets.highcharts import Highchart


class PlotConfigWidget(QWidget, gui.OWComponent):
    sigClosed = pyqtSignal(str, QWidget)
    sigLogarithmic = pyqtSignal(str, bool)
    sigType = pyqtSignal(str, str)
    sigSelection = pyqtSignal(str, list)

    is_logarithmic = False
    plot_type = 'line'

    def __init__(self, owwidget, ax, varmodel):
        QWidget.__init__(self, owwidget)
        gui.OWComponent.__init__(self)

        self.ax = ax
        self.view = view = QListView(
            self, selectionMode=QTreeWidget.ExtendedSelection,
        )
        view.setModel(varmodel)
        selection = view.selectionModel()
        selection.selectionChanged.connect(self.selection_changed)

        box = QVBoxLayout(self)
        box.setContentsMargins(0, 0, 0, 0)
        self.setLayout(box)

        hbox = gui.hBox(self)
        gui.comboBox(hbox, self, 'plot_type',
                     label='Type:',
                     orientation='horizontal',
                     items=('line', 'step line', 'column', 'area', 'spline'),
                     sendSelectedValue=True,
                     callback=lambda: self.sigType.emit(ax, self.plot_type))
        gui.rubber(hbox)
        self.button_close = button = QPushButton('×', hbox,
                                                 visible=False,
                                                 minimumSize=QSize(20, 20),
                                                 maximumSize=QSize(20, 20),
                                                 styleSheet='''
                                                     QPushButton {
                                                         font-weight: bold;
                                                         font-size:14pt;
                                                         margin:0;
                                                         padding:0;
                                                     }''')
        button.clicked.connect(lambda: self.sigClosed.emit(ax, self))
        hbox.layout().addWidget(button)
        gui.checkBox(
            self,
            self,
            "is_logarithmic",
            "Logarithmic axis",
            callback=self.on_logarithmic,
        )
        box.addWidget(view)

    # This is here because sometimes enterEvent/leaveEvent below were called
    # before the constructor set button_close appropriately. I have no idea.
    button_close = None

    def enterEvent(self, event):
        if self.button_close:
            self.button_close.setVisible(True)

    def leaveEvent(self, event):
        if self.button_close:
            self.button_close.setVisible(False)

    def on_logarithmic(self) -> None:
        """
        Callback when changes the the scale to logarithmic or back
        """
        self.sigLogarithmic.emit(self.ax, self.is_logarithmic)

    def selection_changed(self):
        selection = self.get_selection()
        self.sigSelection.emit(self.ax, selection)
        self.sigType.emit(self.ax, self.plot_type)

    def get_selection(self) -> List[Variable]:
        """
        Get variables selected in the view.

        Returns
        -------
        List with selected variables
        """
        return [
            mi.model()[mi.row()]
            for mi in self.view.selectionModel().selectedIndexes()
        ]

    def set_selection(self, values: List[Variable]) -> None:
        """
        Select variables in list and update the graph.

        Parameters
        ----------
        values
            Variables to select
        """
        sel_model = self.view.selectionModel()
        sel_model.clearSelection()
        model = self.view.model()
        for v in values:
            index = model.indexOf(v)
            index = model.index(index, 0)
            sel_model.select(index, sel_model.Select)

    def set_logarithmic(self, is_log: bool) -> None:
        """
        Set is logarithmic setting and update the graph.

        Parameters
        ----------
        is_log
            Boolean that indicates whether to set scale to logarithmic.
        """
        self.is_logarithmic = is_log
        self.on_logarithmic()


class Highstock(Highchart):

    def __init__(self, parent, *args, **kwargs):
        super().__init__(parent, *args,
                         yAxis_lineWidth=2,
                         yAxis_labels_x=6,
                         yAxis_labels_y=-3,
                         yAxis_labels_align_x='right',
                         yAxis_title_text=None,
                         enable_scrollbar = False,
                         plotOptions_series_dataGrouping_groupPixelWidth=2,
                         plotOptions_series_dataGrouping_approximation='high',
                         plotOptions_areasplinerange_states_hover_lineWidthPlus=0,
                         plotOptions_areasplinerange_tooltip_pointFormat='''
                            <span style="color:{point.color}">\u25CF</span>
                            {series.name}: <b>{point.low:.2f} – {point.high:.2f}</b><br/>''',
                         **kwargs)
        self.parent = parent
        self.axes = []

    def _resizeAxes(self):
        if not self.axes:
            return
        MARGIN = 2
        HEIGHT = (100 - (len(self.axes) - 1) * MARGIN) // len(self.axes)
        self.evalJS('''
            var SKIP_AXES = 1,
                HEIGHT = %(HEIGHT)f,
                MARGIN = %(MARGIN)f;
            for (var i = 0; i < chart.yAxis.length - SKIP_AXES; ++i) {
                var top_offset = i * (HEIGHT + MARGIN);
                chart.yAxis[i + SKIP_AXES].update({
                    top: top_offset + '%%',
                    height: HEIGHT + '%%',
                    offset: 0  // Fixes https://github.com/highcharts/highcharts/issues/5199
                }, false);
            }
            chart.reflow();
            chart.redraw(false);
        ''' % locals())

    def addAxis(self):
        from random import random
        ax = 'ax_' + str(random())[2:]
        self.axes.append(ax)
        self.evalJS('''
            chart.addAxis({
                id: '%(ax)s',
            }, false, false, false);
        ''' % locals())
        self._resizeAxes()
        # TODO: multiple series on the bottom navigator, http://jsfiddle.net/highcharts/SD4XN/
        return ax

    def add_legend(self):
        self.evalJS('''chart.legend.options.enabled = true;''')

    def removeAxis(self, ax):
        self.axes.remove(ax)
        self.evalJS('''
            chart.get('%(ax)s').remove();
        ''' % dict(ax=ax))
        self._resizeAxes()

    def setXAxisType(self, ax_type):
        self.evalJS('''
        for (var i=0; i<chart.xAxis.length; ++i) {
            chart.xAxis[i].update({type: '%s'});
        }
        ''' % ax_type)

    def setSeries(self, ax, series):
        """TODO: Clean this shit up"""
        newseries = []
        names = []
        deltas = []
        forecasts = []
        forecasts_ci = []
        ci_percents = []

        data = self.parent.data

        for attr in series:
            newseries.append(data.get_column_view(attr)[0])
            names.append(attr.name)
            deltas.append(None)
            forecasts.append(None)
            forecasts_ci.append(None)
            ci_percents.append(None)

            for forecast in self.parent.forecasts.values():
                fc_attr = attr.name + ' (forecast)'
                if fc_attr in forecast.domain:
                    fc_attr = forecast.domain[fc_attr]
                    # Forecast extends from last known value
                    forecasts[-1] = np.concatenate((newseries[-1][-1:],
                                                    np.ravel(forecast[:, fc_attr])))
                    # ci_attrs = forecast.attributes.get('ci_attrs', {})
                    # ci_low, ci_high = ci_attrs.get(fc_attr, (None, None))
                    ci_low, ci_high = getattr(fc_attr, 'ci_attrs', (None, None))
                    if ci_low in forecast.domain and ci_high in forecast.domain:
                        ci_percents[-1] = ci_low.ci_percent
                        forecasts_ci[-1] = np.row_stack((
                            [newseries[-1][-1]] * 2,  # last known value
                            np.column_stack((
                                np.ravel(forecast[:, ci_low]),
                                np.ravel(forecast[:, ci_high])))))
                    break

            if isinstance(data.time_variable, TimeVariable):
                delta = data.time_delta.backwards_compatible_delta
                tvals = data.time_values

                if isinstance(delta, Number):
                    deltas[-1] = (tvals[0] * 1000, tvals[-1] * 1000, delta * 1000, None)
                elif delta:
                    deltas[-1] = (tvals[0] * 1000, tvals[-1] * 1000) + delta
                else:
                    newseries[-1] = np.column_stack((tvals * 1000,
                                                     newseries[-1]))
                    if forecasts[-1] is not None:
                        if forecast.time_variable:  # Use forecast time if available
                            fc_tvals = np.concatenate((tvals[-1:],
                                                       forecast.time_values * 1000))
                            forecasts[-1] = np.column_stack((
                                fc_tvals, forecasts[-1]))
                            forecasts_ci[-1] = np.column_stack((
                                fc_tvals, forecasts_ci[-1]))
                        else:  # Extrapolate from the final time of data
                            fc_tvals = np.linspace(
                                1000 * tvals[-1],
                                1000 * (tvals[-1] + (len(forecasts[-1]) - 1) * np.diff(tvals[-2:])[0]),
                                len(forecasts[-1]))
                            forecasts[-1] = np.column_stack((
                                fc_tvals, forecasts[-1]))
                            forecasts_ci[-1] = np.column_stack((
                                fc_tvals, forecasts_ci[-1]))

        self.exposeObject('series_' + ax, {'data': newseries,
                                           'forecasts': forecasts,
                                           'forecasts_ci': forecasts_ci,
                                           'ci_percents': ci_percents,
                                           'names': names,
                                           'deltas': deltas})
        self.evalJS('''
            var ax = chart.get('%(ax)s');
            chart.series
            .filter(function(s) { return s.yAxis == ax })
            .map(function(s) { s.remove(false); });

            var data = series_%(ax)s.data,
                names = series_%(ax)s.names,
                deltas = series_%(ax)s.deltas,
                forecasts = series_%(ax)s.forecasts,
                ci_percents = series_%(ax)s.ci_percents,
                forecasts_ci = series_%(ax)s.forecasts_ci;

            for (var i=0; i < data.length; ++i) {
                var opts = {
                    data: data[i],
                    name: names[i],
                    yAxis: '%(ax)s'
                };

                if (deltas[i]) {
                    opts.pointStart = deltas[i][0];
                    // skip 1: pointEnd (forecast start)
                    opts.pointInterval = deltas[i][2];
                    if (deltas[i][3])
                        opts.pointIntervalUnit = deltas[i][3];
                }

                var added_series = chart.addSeries(opts, false, false);

                if (forecasts[i]) {
                    var opts = {
                        linkedTo: ':previous',
                        name: names[i] + ' (forecast)',
                        yAxis: '%(ax)s',
                        data: forecasts[i],
                        dashStyle: 'ShortDash',
                        color: added_series.color,
                        fillOpacity: .3,
                    };
                    if (deltas[i]) {
                        opts.pointStart = deltas[i][1];
                        opts.pointInterval = deltas[i][2];
                        if (deltas[i][3])
                            opts.pointIntervalUnit = deltas[i][3];
                    }
                    chart.addSeries(opts, false, false)
                }
                if (forecasts_ci[i]) {
                    var opts = {
                        type: 'areasplinerange',
                        linkedTo: ':previous',
                        name: names[i] + ' (forecast; ' + ci_percents[i] + '%% CI)',
                        yAxis: '%(ax)s',
                        data: forecasts_ci[i],
                        color: added_series.color,
                        fillOpacity: 0.2,
                        lineWidth: 0,
                    };
                    if (deltas[i]) {
                        opts.pointStart = deltas[i][1];
                        opts.pointInterval = deltas[i][2];
                        if (deltas[i][3])
                            opts.pointIntervalUnit = deltas[i][3];
                    }
                    chart.addSeries(opts, false, false)
                }
            }
            chart.redraw(false);
        ''' % dict(ax=ax))

    def setLogarithmic(self, ax, is_logarithmic):
        self.evalJS('''
            chart.get('%(ax)s').update({ type: '%(type)s', allowNegativeLog:
            %(negative)s, tickInterval: %(tick)s,
        });
        ''' % dict(ax=ax, type='logarithmic' if is_logarithmic else 'linear',
                   negative='true' if is_logarithmic else 'false',
                   tick=1 if is_logarithmic else 'undefined'))
        if not is_logarithmic:
            # it is a workaround for Highcharts issue - Highcharts do not
            # un-mark data as null when changing graph from logarithmic to
            # linear
            self.evalJS(
                '''
                s = chart.get('%(ax)s').series;
                s.forEach(function(series) {
                    series.data.forEach(function(point) {
                        point.update({'isNull': false});
                    });
                });
                ''' % dict(ax=ax)
            )

    def setType(self, ax, type):
        step, type = ('true', 'line') if type == 'step line' else ('false', type)
        self.evalJS('''
            var ax = chart.get('%(ax)s');
            chart.series
            .filter(function(s) { return s.yAxis == ax; })
            .map(function(s) {
                s.update({
                    type: '%(type)s',
                    step: %(step)s
                }, false);
            });
            chart.redraw(false);
        ''' % locals())


class LineChartContextHandler(DomainContextHandler):
    """
    Context handler of line chart. The specifics of this widget is
    that `attrs` variable is list of lists and it is not handled with
    the DomainContextHandler.
    """

    def encode_setting(self, context, setting, value):
        if setting.name == "attrs":
            value = [
                [(var.name, vartype(var)) for var in plot] for plot in value
            ]
        return super().encode_setting(context, setting, value)

    def decode_setting(self, setting, value, domain=None, *_args):
        decoded = super().decode_setting(setting, value, domain)
        if setting.name == "attrs":
            decoded = [[domain[name] for name, _ in plot] for plot in decoded]
        return decoded

    def match(self, context, _, attrs, metas):
        if context.attributes == attrs and context.metas == metas:
            return self.PERFECT_MATCH

        if "attrs" not in context.values:
            return self.NO_MATCH

        all_vars = attrs.copy()
        all_vars.update(metas)

        value = set(y for x in context.values["attrs"] for y in x)
        return sum(all_vars.get(attr) == vtype for attr, vtype in value) / len(
            value
        )


class OWLineChart(OWWidget):
    name = 'Line Chart'
    description = "Visualize time series' sequence and progression."
    icon = 'icons/LineChart.svg'
    priority = 90

    class Inputs:
        time_series = Input("Time series", Table)
        features = Input("Features", AttributeList)
        forecast = Input("Forecast", Timeseries, multiple=True)

    settingsHandler = LineChartContextHandler()
    attrs = ContextSetting([])  # Maps data.name -> [attrs]
    is_logit = ContextSetting([])

    graph_name = 'chart'

    def __init__(self):
        self.data = None
        self.features = None
        self.configs = []
        self.forecasts = OrderedDict()
        self.varmodel = VariableListModel(parent=self)
        icon = QIcon(join(dirname(__file__), 'icons', 'LineChart-plus.png'))
        self.add_button = button = QPushButton(icon, ' &Add plot', self)
        button.clicked.connect(self.add_plot)
        self.controlArea.layout().addWidget(button)
        self.configsArea = gui.vBox(self.controlArea)
        self.controlArea.layout().addStretch(1)
        # TODO: allow selecting ranges that are sent to output as subset table
        self.chart = highstock = Highstock(self, highchart='StockChart')
        self.mainArea.layout().addWidget(highstock)
        # highstock.evalJS('Highcharts.setOptions({navigator: {enabled:false}});')
        highstock.chart(
            # For some reason, these options don't work as global opts applied at Highstock init time
            # Disable top range selector
            rangeSelector_enabled=False,
            rangeSelector_inputEnabled=False,
            # Disable bottom miniview navigator (it doesn't update)
            navigator_enabled=False, )
        self.add_plot()
        self.chart.add_legend()

    def add_plot(self):
        ax = self.chart.addAxis()
        config = PlotConfigWidget(self, ax, self.varmodel)
        # Connect the signals
        config.sigSelection.connect(self.chart.setSeries)
        config.sigLogarithmic.connect(self.chart.setLogarithmic)
        config.sigType.connect(self.chart.setType)
        config.sigClosed.connect(self.chart.removeAxis)
        config.sigClosed.connect(lambda ax, widget: widget.setParent(None))
        config.sigClosed.connect(lambda ax, widget:
                                 self.add_button.setDisabled(False))
        self.configs.append(config)
        self.add_button.setDisabled(len(self.configs) >= 5)
        config.sigClosed.connect(lambda ax, widget: self.remove_plot(widget))

        self.configsArea.layout().addWidget(config)

    def remove_plot(self, plot):
        self.configs.remove(plot)
        # # self.configsArea.layout()
        if len(self.chart.axes) < 2:
            self.resize(QSize(925, 635))

    @Inputs.time_series
    def set_data(self, data):
        # TODO: set xAxis resolution and tooltip time contents depending on
        # data.time_delta. See: http://imgur.com/yrnlgQz

        # If the same data is updated, short circuit to just updating the chart
        # retaining all panels and list view selections ...
        new_data = None if data is None else \
                   Timeseries.from_data_table(data)
        if new_data is not None and self.data is not None \
                and new_data.domain == self.data.domain:
            self.data = new_data
            self._selections_changed()
            return

        self.data = data = new_data
        if data is None:
            self.varmodel.clear()
            self.chart.clear()
            return

        self.set_attributes()

        if getattr(data.time_variable, 'utc_offset', False):
            offset_minutes = data.time_variable.utc_offset.total_seconds() / 60
            self.chart.evalJS(
                'Highcharts.setOptions({global: {timezoneOffset: %d}});'
                % -offset_minutes
            )  # Why is this negative? It works.
            self.chart.chart()

        self.chart.setXAxisType(
            'datetime'
            if (data.time_variable and
                (getattr(data.time_variable, 'have_date', False) or
                 getattr(data.time_variable, 'have_time', False))) else
            'linear')

        variables = [
            var
            for var in data.domain.variables
            if var.is_continuous and var != data.time_variable
        ]
        self.varmodel.wrap(variables)
        self.update_plots()

    @Inputs.forecast
    def set_forecast(self, forecast, id):
        if forecast is not None:
            self.forecasts[id] = forecast
        else:
            self.forecasts.pop(id, None)
        # TODO: update currently shown plots

    @Inputs.features
    def set_features(self, features: AttributeList) -> None:
        if features and not self.features:
            # if features are on the input and they were not before
            # context should be saved for later when features will not be
            # present anymore
            self.closeContext()
        self.features = features
        if self.data:
            self.set_attributes()
            self.update_plots()

    def set_attributes(self) -> None:
        """
        In case when features present: set shown attributes to match features
        In case when features not present: set default value and open context
        """
        self.closeContext()
        if self.features:
            self.attrs = [
                [f]
                for f in self.features
                if f in self.data.domain and f != self.data.time_variable
            ]
            self.is_logit = [False] * len(self.attrs)
        else:
            variables = [
                var
                for var in self.data.domain.variables
                if var.is_continuous and var != self.data.time_variable
            ]
            self.attrs = [[variables[0]]]
            self.is_logit = [False]
            # context is only open when features not provided on input
            # when provided features defines the selection
            self.openContext(self.data.domain)

    def update_plots(self) -> None:
        """
        Update plots when new data or new selection comes
        """
        # remove plots if too many of them
        while len(self.configs) > len(self.attrs):
            plot = self.configs[-1]
            plot.sigClosed.emit(plot.ax, plot)

        # add plots if not enough of them
        while len(self.configs) < len(self.attrs):
            self.add_plot()

        assert len(self.configs) == len(self.attrs)
        assert len(self.attrs) == len(self.is_logit)
        # select correct values
        for config, attr, log in zip(self.configs, self.attrs, self.is_logit):
            config.set_logarithmic(log)
            config.set_selection(attr)
            config.view.setEnabled(not self.features)

    def _selections_changed(self) -> None:
        for config in self.configs:
            config.selection_changed()

    def closeEvent(self, event: QCloseEvent) -> None:
        """
        When someone delets the widget closeContext must be called to gather
        settings.
        """
        self.closeContext()
        super().closeEvent(event)

    def closeContext(self) -> None:
        """
        Gather configs in contextVariables and close context.
        """
        if not self.features:
            # only close in case of when features are not present if they are
            # feature selection is defined by the input and context should
            # not have impact
            attrs, is_logit = [], []
            for config in self.configs:
                attrs.append(config.get_selection())
                is_logit.append(config.is_logarithmic)
            self.attrs = attrs
            self.is_logit = is_logit
            super().closeContext()


if __name__ == "__main__":
    from orangecontrib.timeseries import ARIMA, VAR

    airpassengers = Timeseries.from_file('airpassengers')
    msft = airpassengers.interp()
    model1 = ARIMA((3, 1, 1)).fit(airpassengers)
    model2 = VAR(4).fit(msft)
    ow = WidgetPreview(OWLineChart)
    ow.run(
        set_data=airpassengers,
        set_forecast=[
            (model1.predict(10, as_table=True), 0),
            (model2.predict(10, as_table=True), 1)
        ]
    )



from datetime import datetime, timezone
from itertools import product
import unittest
from unittest.mock import patch

from Orange.data import Table

from orangecontrib.timeseries import Timeseries
from orangecontrib.timeseries.functions import timestamp, fromtimestamp


class TestTimeseries(unittest.TestCase):
    def test_create_time_variable(self):
        table = Table("iris")
        time_series = Timeseries.from_data_table(table)
        id_1 = id(time_series.attributes)
        time_series.time_variable = time_series.domain.attributes[0]
        self.assertNotEqual(id_1, id(time_series.attributes))

    def test_make_timeseries_from_continuous_var(self):
        table = Table.from_url("http://file.biolab.si/datasets/slovenian-national-assembly-eng.tab")
        time_series = Timeseries.make_timeseries_from_continuous_var(table,
                                                                     'date of birth')
        self.assertEqual(time_series.time_variable.name, 'date of birth')
        self.assertTrue(time_series.time_variable in time_series.domain.metas)

    def test_time_var_removed(self):
        ts_with_tv = Timeseries.from_file('airpassengers')
        # select columns without time variable
        ts_without_tv = Timeseries.from_data_table(ts_with_tv[:,
                                      ts_with_tv.domain.class_var])
        self.assertTrue(ts_with_tv.time_variable)
        # make sure the Timeseries without time variable in domain has
        # time_variable set to None
        self.assertIsNone(ts_without_tv.time_variable)


class TestTimestamp(unittest.TestCase):
    def test_timestamp(self):
        local = datetime.now().astimezone().tzinfo

        class T(datetime):
            def timestamp(self):
                nonlocal was_hit
                was_hit = True
                raise OverflowError

        # test different years since 1900 was not a leap year, 2000 was
        # test different timezones to account for naive and aware datetime
        for y, tz in product([1890, 1991, 2004], [None, timezone.utc, local]):
            was_hit = False
            date = datetime(y, 5, 1, 19, 20, 21, 5, tzinfo=tz)
            test_date = T(y, 5, 1, 19, 20, 21, 5, tzinfo=tz)

            self.assertEqual(date.timestamp(), timestamp(test_date))
            self.assertTrue(was_hit)

    def test_fromtimestamp(self):
        TS = -1234567890
        expected = datetime.fromtimestamp(TS)

        was_hit = False

        class MockDatetime(datetime):
            @classmethod
            def fromtimestamp(cls, *args, **kwargs):
                nonlocal was_hit
                if was_hit:
                    return super().fromtimestamp(*args, **kwargs)
                was_hit = True
                raise OSError

        with patch('datetime.datetime', MockDatetime):
            self.assertEqual(fromtimestamp(TS), expected)
            self.assertTrue(was_hit)



from enum import Enum
from itertools import chain
from numbers import Number
from collections import defaultdict
from os import path

import numpy as np

from Orange.data import Table, TimeVariable, DiscreteVariable
from Orange.util import color_to_hex
from Orange.widgets import widget, gui, settings
from Orange.widgets.utils.colorpalettes import ContinuousPalette
from Orange.widgets.utils.itemmodels import VariableListModel
from orangewidget.utils.widgetpreview import WidgetPreview
from Orange.widgets.widget import Input, Output
from orangecontrib.timeseries import Timeseries, fromtimestamp
from orangecontrib.timeseries.agg_funcs import AGG_OPTIONS, Mode
from orangecontrib.timeseries.widgets.highcharts import Highchart

red_palette = ContinuousPalette('Linear Red', 'linear_red',
                                [[204, 0, 0], [204, 1, 0], [204, 1, 1], [204, 2, 2], [204, 3, 3], [205, 4, 4], [205, 5, 4], [205, 5, 5], [205, 6, 6], [205, 7, 7], [206, 8, 8], [206, 9, 8], [206, 9, 9], [206, 10, 10], [206, 11, 11], [207, 12, 12], [207, 13, 12], [207, 13, 13], [207, 14, 14], [207, 15, 15], [208, 16, 16], [208, 17, 16], [208, 17, 17], [208, 18, 18], [208, 19, 19], [209, 20, 20], [209, 21, 20], [209, 21, 21], [209, 22, 22], [209, 23, 23], [210, 24, 24], [210, 25, 24], [210, 25, 25], [210, 26, 26], [210, 27, 27], [211, 28, 28], [211, 29, 28], [211, 29, 29], [211, 30, 30], [211, 31, 31], [212, 32, 32], [212, 33, 32], [212, 33, 33], [212, 34, 34], [212, 35, 35], [213, 36, 36], [213, 37, 36], [213, 37, 37], [213, 38, 38], [213, 39, 39], [214, 40, 40], [214, 41, 40], [214, 41, 41], [214, 42, 42], [214, 43, 43], [215, 44, 44], [215, 45, 44], [215, 45, 45], [215, 46, 46], [215, 47, 47], [216, 48, 48], [216, 49, 48], [216, 49, 49], [216, 50, 50], [216, 51, 51], [217, 52, 52], [217, 53, 52], [217, 53, 53], [217, 54, 54], [217, 55, 55], [218, 56, 56], [218, 57, 56], [218, 57, 57], [218, 58, 58], [218, 59, 59], [219, 60, 60], [219, 61, 60], [219, 61, 61], [219, 62, 62], [219, 63, 63], [220, 64, 64], [220, 65, 64], [220, 65, 65], [220, 66, 66], [220, 67, 67], [221, 68, 68], [221, 69, 68], [221, 69, 69], [221, 70, 70], [221, 71, 71], [222, 72, 72], [222, 73, 72], [222, 73, 73], [222, 74, 74], [222, 75, 75], [223, 76, 76], [223, 77, 76], [223, 77, 77], [223, 78, 78], [223, 79, 79], [224, 80, 80], [224, 81, 80], [224, 81, 81], [224, 82, 82], [224, 83, 83], [225, 84, 84], [225, 85, 84], [225, 85, 85], [225, 86, 86], [225, 87, 87], [226, 88, 88], [226, 89, 88], [226, 89, 89], [226, 90, 90], [226, 91, 91], [227, 92, 92], [227, 93, 92], [227, 93, 93], [227, 94, 94], [227, 95, 95], [228, 96, 96], [228, 97, 96], [228, 97, 97], [228, 98, 98], [228, 99, 99], [229, 100, 100], [229, 101, 100], [229, 101, 101], [229, 102, 102], [229, 103, 103], [230, 104, 104], [230, 105, 104], [230, 105, 105], [230, 106, 106], [230, 107, 107], [231, 108, 108], [231, 109, 108], [231, 109, 109], [231, 110, 110], [231, 111, 111], [232, 112, 112], [232, 113, 112], [232, 113, 113], [232, 114, 114], [232, 115, 115], [233, 116, 116], [233, 117, 116], [233, 117, 117], [233, 118, 118], [233, 119, 119], [234, 120, 120], [234, 121, 120], [234, 121, 121], [234, 122, 122], [234, 123, 123], [235, 124, 124], [235, 125, 124], [235, 125, 125], [235, 126, 126], [235, 127, 127], [236, 128, 128], [236, 129, 128], [236, 129, 129], [236, 130, 130], [236, 131, 131], [237, 132, 132], [237, 133, 132], [237, 133, 133], [237, 134, 134], [237, 135, 135], [238, 136, 136], [238, 137, 136], [238, 137, 137], [238, 138, 138], [238, 139, 139], [239, 140, 140], [239, 141, 140], [239, 141, 141], [239, 142, 142], [239, 143, 143], [240, 144, 144], [240, 145, 144], [240, 145, 145], [240, 146, 146], [240, 147, 147], [241, 148, 148], [241, 149, 148], [241, 149, 149], [241, 150, 150], [241, 151, 151], [242, 152, 152], [242, 153, 152], [242, 153, 153], [242, 154, 154], [242, 155, 155], [243, 156, 156], [243, 157, 156], [243, 157, 157], [243, 158, 158], [243, 159, 159], [244, 160, 160], [244, 161, 160], [244, 161, 161], [244, 162, 162], [244, 163, 163], [245, 164, 164], [245, 165, 164], [245, 165, 165], [245, 166, 166], [245, 167, 167], [246, 168, 168], [246, 169, 168], [246, 169, 169], [246, 170, 170], [246, 171, 171], [247, 172, 172], [247, 173, 172], [247, 173, 173], [247, 174, 174], [247, 175, 175], [248, 176, 176], [248, 177, 176], [248, 177, 177], [248, 178, 178], [248, 179, 179], [249, 180, 180], [249, 181, 180], [249, 181, 181], [249, 182, 182], [249, 183, 183], [250, 184, 184], [250, 185, 184], [250, 185, 185], [250, 186, 186], [250, 187, 187], [251, 188, 188], [251, 189, 188], [251, 189, 189], [251, 190, 190], [251, 191, 191], [252, 192, 192], [252, 193, 192], [252, 193, 193], [252, 194, 194], [252, 195, 195], [253, 196, 196], [253, 197, 196], [253, 197, 197], [253, 198, 198], [253, 199, 199], [254, 200, 200], [254, 201, 200], [254, 201, 201], [254, 202, 202], [254, 203, 203], [255, 204, 204]]
                                )


class Spiralogram(Highchart):
    """
    A radial heatmap.

    Fiddle with it: https://jsfiddle.net/4v87fo2q/5/
    https://jsfiddle.net/avxg2za9/1/
    """

    class AxesCategories(Enum):
        YEARS = ('', lambda _, d: d.year)
        MONTHS = ('', lambda _, d: d.month)
        DAYS = ('', lambda _, d: d.day)
        MONTHS_OF_YEAR = (tuple(range(1, 13)), lambda _, d: d.month)
        DAYS_OF_WEEK = (tuple(range(0, 7)), lambda _, d: d.weekday())
        DAYS_OF_MONTH = (tuple(range(1, 32)), lambda _, d: d.day)
        DAYS_OF_YEAR = (
        tuple(range(1, 367)), lambda _, d: d.timetuple().tm_yday)
        WEEKS_OF_YEAR = (tuple(range(1, 54)), lambda _, d: d.isocalendar()[1])
        WEEKS_OF_MONTH = (tuple(range(1, 6)), lambda _, d: int(
            np.ceil((d.day + d.replace(day=1).weekday()) / 7)))
        HOURS_OF_DAY = (tuple(range(24)), lambda _, d: d.hour)
        MINUTES_OF_HOUR = (tuple(range(60)), lambda _, d: d.minute)

        @staticmethod
        def month_name(month):
            return ('January', 'February', 'March', 'April', 'May', 'June',
                    'July', 'August', 'September', 'October', 'November',
                    'December')[month - 1]

        @staticmethod
        def weekday_name(weekday):
            return (
            'Monday', 'Tuesday', 'Wednesday', 'Thursday', 'Friday', 'Saturday',
            'Sunday')[weekday]

        @classmethod
        def name_it(cls, dim):
            if dim == cls.MONTHS_OF_YEAR:
                return lambda val: cls.month_name(val)
            if dim == cls.DAYS_OF_WEEK:
                return lambda val: cls.weekday_name(val)
            return lambda val: val

    def setSeries(self, timeseries, attr, xdim, ydim, fagg):
        if timeseries is None or not attr:
            self.clear()
            return
        if isinstance(xdim, str) and xdim.isdigit():
            xdim = [str(i) for i in range(1, int(xdim) + 1)]
        if isinstance(ydim, str) and ydim.isdigit():
            ydim = [str(i) for i in range(1, int(ydim) + 1)]

        if isinstance(xdim, DiscreteVariable):
            xcol = timeseries.get_column_view(xdim)[0]
            xvals, xfunc = xdim.values, lambda i, _: xdim.repr_val(xcol[i])
        else:
            xvals, xfunc = xdim.value
        if isinstance(ydim, DiscreteVariable):
            ycol = timeseries.get_column_view(ydim)[0]
            yvals, yfunc = ydim.values, lambda i, _: ydim.repr_val(ycol[i])
        else:
            yvals, yfunc = ydim.value

        values = timeseries.get_column_view(attr)[0]
        time_values = [fromtimestamp(i, tz=timeseries.time_variable.timezone)
                       for i in timeseries.time_values]

        if not yvals:
            yvals = sorted(set(yfunc(i, v) for i, v in enumerate(time_values) if
                               v is not None))
        if not xvals:
            xvals = sorted(set(xfunc(i, v) for i, v in enumerate(time_values) if
                               v is not None))

        indices = defaultdict(list)
        for i, tval in enumerate(time_values):
            if tval is not None:
                indices[(xfunc(i, tval), yfunc(i, tval))].append(i)

        if self._owwidget.invert_date_order:
            yvals = yvals[::-1]

        series = []
        aggvals = []
        self.indices = []
        xname = self.AxesCategories.name_it(xdim)
        yname = self.AxesCategories.name_it(ydim)
        for yval in yvals:
            data = []
            series.append(dict(name=yname(yval), data=data))
            self.indices.append([])
            for xval in xvals:
                inds = indices.get((xval, yval), ())
                self.indices[-1].append(inds)
                point = dict(y=1)
                data.append(point)
                if inds:
                    try:
                        aggval = np.round(fagg(values[inds]), 4)
                    except ValueError:
                        aggval = np.nan
                else:
                    aggval = np.nan
                if isinstance(aggval, Number) and np.isnan(aggval):
                    aggval = 'N/A'
                    point['select'] = ''
                    point['color'] = 'white'
                else:
                    aggvals.append(aggval)
                point['n'] = aggval

        # TODO: allow scaling over just rows or cols instead of all values as currently
        try:
            maxval, minval = np.max(aggvals), np.min(aggvals)
        except ValueError:
            self.clear()
            return
        ptpval = maxval - minval
        color = red_palette
        for serie in series:
            for point in serie['data']:
                n = point['n']
                if isinstance(n, Number):
                    val = (n - minval) / ptpval

                    if attr.is_discrete and fagg == Mode:
                        point['n'] = attr.repr_val(n)
                    elif isinstance(attr, TimeVariable):
                        point['n'] = attr.repr_val(n)

                    point['color'] = color_to_hex(attr.colors[int(n)]) if \
                        attr.is_discrete else color[val]
                    point['states'] = dict(select=dict(borderColor="black"))

        # TODO: make a white hole in the middle. Center w/o data.
        self.chart(series=series,
                   xAxis_categories=[xname(i) for i in xvals],
                   yAxis_categories=[yname(i) for i in reversed(yvals)],
                   javascript_after='''
                       // Force zoomType which is by default disabled for polar charts
                       chart.options.chart.zoomType = 'xy';
                       chart.pointer.init(chart, chart.options);
                   ''')

    def selection_indices(self, indices):
        result = []
        for i, inds in enumerate(indices):
            if len(inds):
                for j in inds:
                    result.append(self.indices[i][j])
        return sorted(chain.from_iterable(result))

    OPTIONS = dict(
        chart=dict(
            type='column',
            polar=True,
            panning=False, # Fixes: https://github.com/highcharts/highcharts/issues/5240
            events=dict(
                selection='/**/ zoomSelection',  # from _spiralogram.js
            ),
            zoomType='xy',
            # polar=True disabled this, but is again reenabled in JS after chart init
        ),
        legend=dict(
            enabled=False,  # FIXME: Have a heatmap-style legend
        ),
        xAxis=dict(
            gridLineWidth=0,
            showLastLabel=False,
            # categories=None,  # Override this
        ),
        yAxis=dict(
            gridLineWidth=0,
            endOnTick=False,
            showLastLabel=False,
            # categories=None,  # Override this
            labels=dict(
                y=0,
                align='center',
                style=dict(
                    color='black',
                    fontWeight='bold',
                    textShadow=('2px  2px 1px white, -2px  2px 1px white,'
                                '2px -2px 1px white, -2px -2px 1px white'),
                ),
            ),
        ),
        plotOptions=dict(
            column=dict(
                colorByPoint=True,
                stacking='normal',
                pointPadding=0,
                groupPadding=0,
                borderWidth=2,
                pointPlacement='on',
                allowPointSelect=True,
                states=dict(
                    select=dict(
                        borderColor=None,  # Revert Orange's theme
                    )
                )
            )
        ),
        tooltip=dict(
            shared=False,
            formatter=('''/**/
                (function() {
                    if (this.point.n == "N/A")
                        return false;
                    return '<span style="font-size:13pt;color:' + \
                           this.point.color + '">\u25A0</span> ' + \
                           this.series.name + ', ' + \
                           this.x + ': <b>' + \
                           this.point.n + '</b><br/>';
                })
            '''),
        )
        # series=[]  # Override this
    )

    def __init__(self, parent, *args, **kwargs):
        # TODO: Add colorAxes (heatmap legend)
        with open(path.join(path.dirname(__file__), '_spiralogram.js')) as f:
            javascript = f.read()
        super().__init__(parent, *args,
                         options=self.OPTIONS,
                         enable_select='+',  # TODO: implement mouse-drag select
                         javascript=javascript,
                         **kwargs)
        self.indices = {}
        assert isinstance(parent, widget.OWWidget)
        self._owwidget = parent


def _enum_str(enum_value, inverse=False):
    if isinstance(enum_value, DiscreteVariable):
        enum_value = str(enum_value)
    if inverse:
        return enum_value.replace(' ', '_').upper()
    return enum_value.name.replace('_', ' ').lower()


DEFAULT_AGG_FUNC = next(iter(AGG_OPTIONS.keys()))


class OWSpiralogram(widget.OWWidget):
    name = 'Spiralogram'
    description = "Visualize time series' periodicity in a spiral heatmap."
    icon = 'icons/Spiralogram.svg'
    priority = 120

    class Inputs:
        time_series = Input("Time series", Table)

    class Outputs:
        time_series = Output("Time series", Timeseries)

    settings_version = 2
    settingsHandler = settings.DomainContextHandler()

    ax1 = settings.ContextSetting('months of year')
    ax2 = settings.ContextSetting('years')

    agg_attr = settings.ContextSetting(None)
    agg_func = settings.ContextSetting(DEFAULT_AGG_FUNC)

    invert_date_order = settings.Setting(False)

    graph_name = 'chart'

    class Error(widget.OWWidget.Error):
        no_time_variable = widget.Msg(
            'Spiralogram requires time series with a time variable.')

    def __init__(self):
        self.data = None
        self.indices = []
        box = gui.vBox(self.controlArea, 'Axes')
        self.combo_ax2_model = VariableListModel(parent=self)
        self.combo_ax1_model = VariableListModel(parent=self)
        for model in (self.combo_ax1_model, self.combo_ax2_model):
            model[:] = [_enum_str(i) for i in Spiralogram.AxesCategories]
        self.combo_ax2 = gui.comboBox(
            box, self, 'ax2', label='Y axis:', callback=self.replot,
            sendSelectedValue=True, orientation='horizontal',
            model=self.combo_ax2_model)
        self.combo_ax1 = gui.comboBox(
            box, self, 'ax1', label='Radial:', callback=self.replot,
            sendSelectedValue=True, orientation='horizontal',
            model=self.combo_ax1_model)
        gui.checkBox(box, self, 'invert_date_order', 'Invert Y axis order',
                     callback=self.replot)

        box = gui.vBox(self.controlArea, 'Aggregation')

        self.attrs_model = VariableListModel()
        self.attr_cb = gui.comboBox(box, self, 'agg_attr',
                                    sendSelectedValue=True,
                                    model=self.attrs_model,
                                    callback=self.update_agg_combo)

        self.combo_func = gui.comboBox(
            box, self, 'agg_func', label='Function:',
            items=[DEFAULT_AGG_FUNC], orientation='horizontal',
            sendSelectedValue=True,
            callback=self.replot)

        gui.rubber(self.controlArea)

        self.chart = chart = Spiralogram(self,
                                         selection_callback=self.on_selection)
        self.mainArea.layout().addWidget(chart)

    @Inputs.time_series
    def set_data(self, data):
        self.Error.clear()
        self.data = data = None if data is None else \
                           Timeseries.from_data_table(data)

        if data is None:
            self.commit()
            return

        if self.data.time_variable is None or not isinstance(
                self.data.time_variable, TimeVariable):
            self.Error.no_time_variable()
            self.commit()
            return

        def init_combos():
            for model in (self.combo_ax1_model, self.combo_ax2_model):
                model.clear()
            variables = []
            if data is not None and data.time_variable is not None:
                for model in (self.combo_ax1_model, self.combo_ax2_model):
                    model[:] = [_enum_str(i) for i in
                                Spiralogram.AxesCategories]
            for var in data.domain.variables if data is not None else []:
                if (var.is_primitive() and
                        (var is not data.time_variable or
                         isinstance(var, TimeVariable)
                         and data.time_delta.backwards_compatible_delta is None)):
                    variables.append(var)

                if var.is_discrete:
                    for model in (self.combo_ax1_model, self.combo_ax2_model):
                        model.append(var)
            self.attrs_model[:] = variables

        init_combos()
        self.chart.clear()

        self.closeContext()
        self.ax2 = next((self.combo_ax2.itemText(i)
                         for i in range(self.combo_ax2.count())), '')
        self.ax1 = next((self.combo_ax1.itemText(i)
                         for i in range(1, self.combo_ax1.count())), self.ax2)
        self.agg_attr = data.domain[self.attrs_model[0]] if len(
            data.domain.variables) else None
        self.agg_func = DEFAULT_AGG_FUNC

        if getattr(data, 'time_variable', None) is not None:
            self.openContext(data.domain)

        self.update_agg_combo()
        self.replot()

    def update_agg_combo(self):
        self.combo_func.clear()
        new_aggs = AGG_OPTIONS

        if self.agg_attr is not None:
            if self.agg_attr.is_discrete:
                new_aggs = [agg for agg in AGG_OPTIONS if AGG_OPTIONS[agg].disc]
            elif self.agg_attr.is_time:
                new_aggs = [agg for agg in AGG_OPTIONS if AGG_OPTIONS[agg].time]
        self.combo_func.addItems(new_aggs)

        if self.agg_func not in new_aggs:
            self.agg_func = next(iter(new_aggs))

        self.replot()

    def replot(self):
        if not self.combo_ax1.count() or not self.agg_attr:
            return self.chart.clear()

        func = AGG_OPTIONS[self.agg_func].transform
        try:
            ax1 = Spiralogram.AxesCategories[_enum_str(self.ax1, True)]
        except KeyError:
            ax1 = self.data.domain[self.ax1]
        # TODO: Allow having only a single (i.e. radial) axis
        try:
            ax2 = Spiralogram.AxesCategories[_enum_str(self.ax2, True)]
        except KeyError:
            ax2 = self.data.domain[self.ax2]
        self.chart.setSeries(self.data, self.agg_attr, ax1, ax2, func)

    def on_selection(self, indices):
        self.indices = self.chart.selection_indices(indices)
        self.commit()

    def commit(self):
        self.Outputs.time_series.send(
            self.data[self.indices] if self.data else None)

    @classmethod
    def migrate_context(cls, context, version):
        if version < 2:
            values = context.values
            context.values["agg_attr"] = values["agg_attr"][0][0]
            _, type = values["agg_attr"]
            ind, pos = values["agg_func"]
            if type == 101: # discrete variable is always Mode in old settings
                context.values["agg_func"] = ('Mode', pos)
            else:
                context.values["agg_func"] = (list(AGG_OPTIONS)[ind], pos)


if __name__ == "__main__":
    WidgetPreview(OWSpiralogram).run(Table.from_file('airpassengers'))



from Orange.widgets import widget, gui, settings
from Orange.widgets.widget import Input

from orangecontrib.timeseries import Timeseries, ARIMA
from orangecontrib.timeseries.widgets._owmodel import OWBaseModel


class OWARIMAModel(OWBaseModel):
    name = 'ARIMA Model'
    description = 'Model the time series using ARMA, ARIMA, or ARIMAX.'
    icon = 'icons/ARIMA.svg'
    priority = 210

    p = settings.Setting(1)
    d = settings.Setting(0)
    q = settings.Setting(0)
    use_exog = settings.Setting(False)

    UserAdviceMessages = [
        widget.Message('ARIMA(0,1,0) equals to a random walk model, which '
                       'is an AR(1) model with the auto-regression coefficient '
                       'equal to 1, and the constant term (i.e. the period-to-period '
                       "change) equal to the series' long-term drift.",
                       'random-walk',
                       widget.Message.Warning)
    ]

    class Inputs(OWBaseModel.Inputs):
        exogenous_data = Input("Exogenous data", Timeseries)

    def __init__(self):
        super().__init__()
        self.exog_data = None

    @Inputs.exogenous_data
    def set_exog_data(self, data):
        self.exog_data = data
        self.update_model()

    def add_main_layout(self):
        box = gui.vBox(self.controlArea, box='Parameters')
        gui.spin(box, self, 'p', 0, 100, label='Auto-regression order (p):',
                 callback=self.apply)
        gui.spin(box, self, 'd', 0, 2, label='Differencing degree (d):',
                 callback=self.apply)
        gui.spin(box, self, 'q', 0, 100, label='Moving average order (q):',
                 callback=self.apply)
        gui.checkBox(box, self, 'use_exog',
                     'Use exogenous (independent) variables (ARMAX)',
                     callback=self.apply)

    def forecast(self, model):
        if self.use_exog and self.exog_data is None:
            return
        return model.predict(self.forecast_steps,
                             exog=self.exog_data,
                             alpha=1 - self.forecast_confint / 100,
                             as_table=True)

    def create_learner(self):
        return ARIMA((self.p, self.d, self.q), self.use_exog)


if __name__ == "__main__":
    from AnyQt.QtWidgets import QApplication
    from Orange.data import Domain

    a = QApplication([])
    ow = OWARIMAModel()

    data = Timeseries.from_file('airpassengers')
    domain = Domain(data.domain.attributes[:-1], data.domain.attributes[-1])
    data = Timeseries.from_numpy(domain, data.X[:, :-1], data.X[:, -1])
    ow.set_data(data)

    ow.show()
    a.exec()



import unittest

from Orange.data import Table
from Orange.widgets.tests.base import WidgetTest
from orangecontrib.timeseries.agg_funcs import AGG_FUNCTIONS

from orangecontrib.timeseries.widgets.owmovingtransform import OWMovingTransform


class TestOWMovingTransform(WidgetTest):
    def setUp(self):
        self.widget = self.create_widget(OWMovingTransform)  # type: OWMovingTransform

    def test_no_transforms_added(self):
        """
        Prevent crashing when no transforms are added.
        GH-42
        """
        w = self.widget
        table = Table("airpassengers")
        self.assertFalse(w.non_overlapping)
        w.controls.non_overlapping.click()
        self.assertTrue(w.non_overlapping)
        self.send_signal(w.Inputs.time_series, table)
        self.assertFalse(w.Warning.no_transforms_added.is_shown())
        w.controls.autocommit.click()
        self.assertTrue(w.Warning.no_transforms_added.is_shown())
        w.add_button.click()
        w.controls.autocommit.click()
        self.assertFalse(w.Warning.no_transforms_added.is_shown())

    def test_store_settings(self):
        w = self.widget
        iris = Table("iris")
        heart = Table("heart_disease")

        transformations = [[iris.domain[1], 3, AGG_FUNCTIONS[2]],
                           [iris.domain[2], 5, AGG_FUNCTIONS[3]]]
        self.send_signal(w.Inputs.time_series, iris)
        w.table_model = transformations[:]

        self.send_signal(w.Inputs.time_series, heart)
        self.assertEqual(list(w.table_model), [])

        self.send_signal(w.Inputs.time_series, iris)
        self.assertEqual(list(w.table_model), transformations)

        self.send_signal(w.Inputs.time_series, None)
        self.assertEqual(list(w.table_model), [])

        self.send_signal(w.Inputs.time_series, iris)
        self.assertEqual(list(w.table_model), transformations)


if __name__ == "__main__":
    unittest.main()
